import numpy as np

from ._base import DecompositionMixin


class VanillaFairletDecomposition(DecompositionMixin):
    def __init__(self, p, q):
        self.fairlets = []
        self.fairlet_centers = []
        self.p = p
        self.q = q

    def balanced(self, p, q, r, b):
        if r == 0 and b == 0:
            return True
        if r == 0 or b == 0:
            return False
        return min(r * 1.0 / b, b * 1.0 / r) >= p * 1.0 / q

    def make_fairlet(self, points, dataset):
        "Adds fairlet to fairlet decomposition, returns median cost"
        self.fairlets.append(points)
        cost_list = [
            sum([np.linalg.norm(dataset[center] - dataset[point]) for point in points])
            for center in points
        ]
        cost, center = min((cost, center) for (center, cost) in enumerate(cost_list))
        self.fairlet_centers.append(points[center])
        return cost

    def fit_transform(self, dataset, group_a, group_b):
        """
        Computes vanilla (p,q)-fairlet decomposition of given points (Lemma 3 in NIPS17 paper).
        Returns cost.
        Input: Balance parameters p,q which are non-negative integers satisfying p<=q and gcd(p,q)=1.
        "blues" and "reds" are sets of points indices with balance at least p/q.
        """
        blues = list(np.where(group_a)[0])
        reds = list(np.where(group_b)[0])
        return self.fairlets, self.fairlet_centers, self.decompose(blues, reds, dataset)

    def decompose(self, blues, reds, dataset):
        p = self.p
        q = self.q
        assert p <= q, "Please use balance parameters in the correct order"
        if len(reds) < len(blues):
            temp = blues
            blues = reds
            reds = temp
        R = len(reds)
        B = len(blues)
        assert self.balanced(p, q, R, B), (
            "Input sets are unbalanced: " + str(R) + "," + str(B)
        )

        if R == 0 and B == 0:
            return 0

        b0 = 0
        r0 = 0
        cost = 0
        while (R - r0) - (B - b0) >= q - p and R - r0 >= q and B - b0 >= p:
            cost += self.make_fairlet(reds[r0 : r0 + q] + blues[b0 : b0 + p], dataset)
            r0 += q
            b0 += p
        if R - r0 + B - b0 >= 1 and R - r0 + B - b0 <= p + q:
            cost += self.make_fairlet(reds[r0:] + blues[b0:], dataset)
            r0 = R
            b0 = B
        elif R - r0 != B - b0 and B - b0 >= p:
            cost += self.make_fairlet(
                reds[r0 : r0 + (R - r0) - (B - b0) + p] + blues[b0 : b0 + p], dataset
            )
            r0 += (R - r0) - (B - b0) + p
            b0 += p
        assert R - r0 == B - b0, "Error in computing fairlet decomposition"
        for i in range(R - r0):
            cost += self.make_fairlet([reds[r0 + i], blues[b0 + i]], dataset)
        return cost



from ._categorical_repairer import CategoricalRepairer
from ._utils import FreedmanDiaconisBinSize as bin_calculator
from ._utils import get_median, make_histogram_bins


class NumericalRepairer:
    def __init__(
        self, feature_to_repair, repair_level, kdd=False, features_to_ignore=[]
    ):
        self.feature_to_repair = feature_to_repair
        self.repair_level = repair_level
        self.kdd = kdd
        self.features_to_ignore = features_to_ignore
        self.categoric_repairer = CategoricalRepairer(
            feature_to_repair=feature_to_repair,
            repair_level=repair_level,
            kdd=kdd,
            features_to_ignore=features_to_ignore,
        )

    def repair(self, data_to_repair):

        # Convert the "feature_to_repair" into a pseudo-categorical feature by
        # applying binning on that column.
        binned_data = [row[:] for row in data_to_repair]
        index_bins = make_histogram_bins(
            bin_calculator, data_to_repair, self.feature_to_repair
        )

        category_medians = {}
        for i, index_bin in enumerate(index_bins):
            bin_name = "BIN_{}".format(
                i
            )  # IE, the "category" to replace numeric values.
            for j in index_bin:
                binned_data[j][self.feature_to_repair] = bin_name
            category_vals = [
                data_to_repair[j][self.feature_to_repair] for j in index_bin
            ]
            category_medians[bin_name] = get_median(category_vals, self.kdd)

        repaired_data = self.categoric_repairer.repair(binned_data)

        # Replace the "feature_to_repair" column with the median numeric value.
        for i in range(len(repaired_data)):
            if self.repair_level > 0:
                rep_category = repaired_data[i][self.feature_to_repair]
                repaired_data[i][self.feature_to_repair] = category_medians[
                    rep_category
                ]
            else:
                repaired_data[i][self.feature_to_repair] = data_to_repair[i][
                    self.feature_to_repair
                ]
        return repaired_data



import numpy as np

from holisticai.bias.mitigation.commons.disparate_impact_remover._numerical_repairer import (
    NumericalRepairer,
)
from holisticai.utils.transformers.bias import BMPreprocessing as BMPre
from holisticai.utils.transformers.bias import SensitiveGroups


class DisparateImpactRemover(BMPre):
    """
    Disparate impact remover edits feature values to increase group fairness
    while preserving rank-ordering within groups.

    References
    ----------
        [1] Feldman, Michael, et al. "Certifying and removing disparate impact."
        proceedings of the 21th ACM SIGKDD international conference on knowledge
        discovery and data mining. 2015.
    """

    def __init__(self, repair_level=1.0):
        """
        Disparate Impact Remover Preprocessing Bias Mitigator

        Description
        -----------
        Initialize Mitigator class

        Parameters
        ----------
        repair_level : float
                Repair amount 0.0 (min) -> 1.0 (max)
        """
        self._assert_parameters(repair_level)
        self.repair_level = repair_level
        self.sensgroup = SensitiveGroups()

    def _assert_parameters(self, repair_level):
        if not 0.0 <= repair_level <= 1.0:
            raise ValueError("'repair_level' must be between 0.0 and 1.0.")

    def repair_data(self, X, group_a, group_b):
        sensitive_features = np.c_[group_a, group_b]
        p_attr = self.sensgroup.fit_transform(
            sensitive_features, convert_numeric=True
        ).to_numpy()
        data = np.c_[p_attr, X].tolist()
        dir = NumericalRepairer(
            feature_to_repair=0, repair_level=self.repair_level, kdd=False
        )
        new_data_matrix_np = dir.repair(data)
        return np.array([np.array(row[1:]) for row in new_data_matrix_np])

    def transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Transform data

        Description
        -----------
        Transform data to a fair representation

        Parameters
        ----------
        X : ndarray
            Input data

        group_a : ndarray
            mask vector

        group_b : ndarray
            mask vector

        Return
        ------
            Self
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        return self.repair_data(X, group_a, group_b)

    def fit(self):
        return self

    def fit_transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        return self.fit().transform(X, group_a, group_b)



from typing import Optional, Union

import numpy as np

from holisticai.utils._validation import _check_valid_y_proba
from holisticai.utils.transformers._transformer_base import BMTransformerBase


class BMPostprocessing(BMTransformerBase):
    """
    Base Post Processing transformer
    """

    BM_NAME = "Postprocessing"

    def _load_data(self, **kargs):
        """Save postprocessing atributes and convert data to standard format parameters."""

        params = {}

        if self._has_valid_argument(kargs, "y_true"):
            y_true = np.array(kargs.get("y_true")).ravel()
            params.update({"y_true": y_true})

        if self._has_valid_argument(kargs, "y_proba"):
            y_proba = np.array(kargs.get("y_proba"))

            _check_valid_y_proba(y_proba=y_proba)

            params.update({"y_proba": y_proba})

            nb_classes = y_proba.shape[1]
            if nb_classes == 2:
                favorable_index = 1
                y_score = np.array(y_proba[:, favorable_index]).ravel()
                params.update({"y_score": y_score})

            y_pred = np.argmax(y_proba, axis=1).ravel()
            params.update({"y_pred": y_pred})

        if self._has_valid_argument(kargs, "y_pred"):
            y_pred = np.array(kargs.get("y_pred")).ravel()
            params.update({"y_pred": y_pred})

        params_to_numpy_format = ["group_a", "group_b", "y_score"]
        for param_name in params_to_numpy_format:
            if self._has_valid_argument(kargs, param_name):
                params.update({param_name: self._to_numpy(kargs, param_name)})

        if self._has_valid_argument(kargs, "X"):
            params.update({"X": self._to_numpy(kargs, "X", ravel=False)})

        if self._has_valid_argument(kargs, "sample_weight"):
            params.update({"sample_weight": self._to_numpy(kargs, "sample_weight")})

        elif "y_true" in locals():
            params.update({"sample_weight": np.ones_like(y_true).astype(np.float64)})

        return params

    @staticmethod
    def _has_valid_argument(kargs, name):
        return (name in kargs) and (kargs[name] is not None)



import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error

from ..utils._validation import _multiclass_checks


def confusion_matrix(y_pred, y_true, classes=None, normalize=None):
    """
    Confusion Matrix.

    Description
    ----------
    This function computes the confusion matrix. The i,jth
    entry is the number of elements with predicted class i
    and true class j.

    Parameters
    ----------
    y_pred : array-like
        Prediction vector (categorical)
    y_true : array-like
        Target vector (categorical)
    classes (optional) : list
        The unique output classes in order
    normalize (optional) : None, 'pred' or 'class'
        According to which of pred or class we normalize

    Returns
    -------
    numpy ndarray
        Confusion Matrix : shape (num_classes, num_classes)

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import confusion_matrix
    >>> y_pred = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2, 0])
    >>> y_true = np.array([0, 1, 1, 0, 1, 0, 2, 1, 2, 1])
    >>> confusion_matrix(y_pred, y_true, classes=[2, 1, 0])
        2    1    0
    2  1.0  1.0  1.0
    1  0.0  3.0  0.0
    0  1.0  1.0  2.0
    """
    # check and coerce inputs
    _, y_pred, y_true, _, classes = _multiclass_checks(
        p_attr=None,
        y_pred=y_pred,
        y_true=y_true,
        groups=None,
        classes=classes,
    )

    # variables
    num_classes = len(classes)
    class_dict = dict(zip(classes, range(num_classes)))

    # initialize the confusion matrix
    confmat = np.zeros((num_classes, num_classes))

    # loop over instances
    for x, y in zip(y_pred, y_true):
        # increment correct entry
        confmat[class_dict[x], class_dict[y]] += 1

    if normalize is None:
        pass

    elif normalize == "pred":
        confmat = confmat / np.sum(confmat, axis=1).reshape(-1, 1)

    elif normalize == "true":
        confmat = confmat / np.sum(confmat, axis=0).reshape(1, -1)

    else:
        raise ValueError('normalize should be one of None, "pred" or "true"')

    return pd.DataFrame(confmat, columns=classes).set_index(np.array(classes))


def classification_efficacy_metrics(y_pred, y_true=None, y_proba=None):
    """
    Classification efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant classification efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    y_pred : array-like
        Predictions vector (binary)
    y_true (optional) : numpy array
        Target vector (binary)
    y_proba (optional) : array-like
        Probability Matrix estimates (num_examples, 2)

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    from sklearn import metrics

    perform = {
        "Accuracy": metrics.accuracy_score,
        "Balanced accuracy": metrics.balanced_accuracy_score,
        "Precision": metrics.precision_score,
        "Recall": metrics.recall_score,
        "F1-Score": metrics.f1_score,
    }

    hyper = {
        "Accuracy": {},
        "Balanced accuracy": {},
        "Precision": {},
        "Recall": {},
        "F1-Score": {},
    }

    soft_perform = {"AUC": metrics.roc_auc_score, "Log Loss": metrics.log_loss}

    soft_hyper = {"AUC": {"average": "micro"}, "Log Loss": {}}

    ref_vals = {
        "Accuracy": 1,
        "Balanced accuracy": 1,
        "Precision": 1,
        "Recall": 1,
        "F1-Score": 1,
        "AUC": 1,
        "Log Loss": 0,
    }
    metrics = [
        [pf, fn(y_pred, y_true, **hyper[pf]), ref_vals[pf]]
        for pf, fn in perform.items()
    ]
    if y_proba is not None:
        opp_metrics = [
            [pf, fn(y_pred, y_true, **soft_hyper[pf]), ref_vals[pf]]
            for pf, fn in soft_perform.items()
        ]
        metrics += opp_metrics

    return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index(
        "Metric"
    )


def clustering_efficacy_metrics(X, y_pred):
    """
    Clustering efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant clustering efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    X : matrix-like
        feature matrix
    y_pred (optional) :  array-like
        Predicted vector

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference

    Examples
    --------
    """
    from sklearn import metrics

    perform = {
        "Silhouette": metrics.silhouette_score,
        "Calinski Harabasz": metrics.calinski_harabasz_score,
        "Davies Bouldin": metrics.davies_bouldin_score,
    }

    hyper = {
        "Silhouette": {"metric": "euclidean"},
        "Calinski Harabasz": {},
        "Davies Bouldin": {},
    }

    ref_vals = {
        "Silhouette": 1,
        "Calinski Harabasz": np.inf,
        "Davies Bouldin": 0,
    }
    metrics = [
        [pf, fn(X, y_pred, **hyper[pf]), ref_vals[pf]] for pf, fn in perform.items()
    ]

    return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index(
        "Metric"
    )


def is_numeric(obj):
    attrs = ["__add__", "__sub__", "__mul__", "__truediv__", "__pow__"]
    return all(hasattr(obj, attr) for attr in attrs)


def multiclassification_efficacy_metrics(
    y_pred, y_true=None, y_proba=None, classes=None, by_class=False
):
    """
    Multiclassification efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant multiclassification efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    y_pred : array-like
        Predictions vector
    y_true (optional) : array-like
        Target vector
    y_proba (optional) : array-like
        Probability Matrix estimates (num_examples, num_classes)

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    from sklearn import metrics

    if classes is None:
        classes = sorted(np.unique(y_pred))
        if y_proba is not None:
            assert all(
                [is_numeric(c) for c in classes]
            ), f"to evaluate y_proba outputs, parameter `classes` must be passed."

    def convert_to_numeric_array(y, classes):
        if len(set(classes).intersection(set(np.unique(y)))) == len(classes):
            y = pd.Series(list(y)).replace(dict(zip(classes, range(len(classes)))))
        return y

    y_pred = convert_to_numeric_array(y_pred, classes)
    y_true = convert_to_numeric_array(y_true, classes)

    n_classes = len(classes)

    perform = {
        "Accuracy": metrics.accuracy_score,
        "Balanced accuracy": metrics.balanced_accuracy_score,
        "Precision": metrics.precision_score,
        "Recall": metrics.recall_score,
        "F1-Score": metrics.f1_score,
    }

    hyper = {
        "Accuracy": {},
        "Balanced accuracy": {},
        "Precision": {"average": "micro"},
        "Recall": {"average": "micro"},
        "F1-Score": {"average": "micro"},
    }

    soft_perform = {"AUC": metrics.roc_auc_score, "Log Loss": metrics.log_loss}

    soft_hyper = {
        "AUC": {"average": "macro", "multi_class": "ovr"},
        "Log Loss": {},
    }

    ref_vals = {
        "Accuracy": 1,
        "Balanced accuracy": 1,
        "Precision": 1,
        "Recall": 1,
        "F1-Score": 1,
        "AUC": 1,
        "Log Loss": 0,
    }

    detailed_hyper = {
        "Precision": {"average": None},
        "Recall": {"average": None},
        "F1-Score": {"average": None},
        "AUC": {"average": None, "multi_class": "ovr"},
    }

    def process_value(metrics_dict, hyper_dict, y_pred, y_true):
        values_list = []
        for pf, fn in metrics_dict.items():
            result = fn(y_true, y_pred, **hyper_dict[pf])
            values_list.append(result)
        return values_list

    def process_metrics(
        values_list, metrics_dict, y_pred, y_true_oh, n_classes, detailed_hyper
    ):
        metrics_list = []
        for value, (pf, fn) in zip(values_list, metrics_dict.items()):
            result = [value]
            if pf in detailed_hyper:
                result += list(fn(y_true_oh, y_pred, **detailed_hyper[pf]))
            else:
                result += [""] * n_classes
            metrics_list.append(result)
        return metrics_list

    metrics_1 = process_value(perform, hyper, y_pred, y_true)
    if by_class:
        metrics_1 = process_metrics(
            metrics_1, perform, y_pred, y_true, n_classes, detailed_hyper
        )
        metrics = [
            [pf] + metric + [ref_vals[pf]]
            for metric, (pf, fn) in zip(metrics_1, perform.items())
        ]
    else:
        metrics = [
            [pf] + [metric] + [ref_vals[pf]]
            for metric, (pf, fn) in zip(metrics_1, perform.items())
        ]

    if y_proba is not None:
        y_true_oh = y_true.replace(dict(zip(classes, range(len(classes)))))
        metrics_2 = process_value(soft_perform, soft_hyper, y_proba, y_true_oh)
        if by_class:
            metrics_2 = process_metrics(
                metrics_2, soft_perform, y_proba, y_true_oh, n_classes, detailed_hyper
            )
            metrics_2 = [
                [pf] + metric + [ref_vals[pf]]
                for metric, (pf, fn) in zip(metrics_2, soft_perform.items())
            ]
        else:
            metrics_2 = [
                [pf] + [metric] + [ref_vals[pf]]
                for metric, (pf, fn) in zip(metrics_2, soft_perform.items())
            ]
        metrics += metrics_2

    values_cols = ["Value"]
    if by_class:
        values_cols += classes

    return pd.DataFrame(
        metrics, columns=["Metric"] + values_cols + ["Reference"]
    ).set_index("Metric")


def rmse_score(y_true, y_pred, **kargs):
    return np.sqrt(mean_squared_error(y_true, y_pred, **kargs))


def regression_efficacy_metrics(y_pred, y_true):
    """
    Regression efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant regression efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    y_pred : array-like
        Predictions vector
    y_true (optional) : array-like
        Target vector

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    from sklearn import metrics

    perform = {
        # "Pseudo-R2": pr2_score,
        "RMSE": rmse_score,
        "MAE": metrics.mean_absolute_error,
        "MAPE": metrics.mean_absolute_percentage_error,
        "Max Error": metrics.max_error,
    }

    hyper = {
        # "Pseudo-R2": {},
        "RMSE": {"sample_weight": None, "multioutput": "uniform_average"},
        "MAE": {"sample_weight": None, "multioutput": "uniform_average"},
        "MAPE": {"sample_weight": None, "multioutput": "uniform_average"},
        "Max Error": {},
    }

    ref_vals = {
        "RMSE": 0,
        "MAE": 0,
        "MAPE": 0,
        "Max Error": 0,
    }

    metrics = [
        [pf, fn(y_pred, y_true, **hyper[pf]), ref_vals[pf]]
        for pf, fn in perform.items()
    ]

    return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index(
        "Metric"
    )



from typing import Optional

import numpy as np
import pandas as pd

from ..commons._conventions import *


class GridGenerator:
    """
    Create a grid using a constraint
    """

    def __init__(
        self,
        grid_size: Optional[int] = 5,
        grid_limit: Optional[int] = 2.0,
        neg_allowed: Optional[pd.DataFrame] = None,
        force_L1_norm: Optional[bool] = True,
    ):
        """
        Initialize Grid Generator

        Parameters
        ----------
        grid_size: int
            number of columns to be generated in the grid.

        grid_limit : float
            range of the values in the grid generated.

        neg_values: bool
            ensures if we want to include negative values in the grid or not.
            If True, the range is doubled.
        """
        self.grid_size = grid_size
        self.grid_limit = grid_limit
        self.neg_allowed = neg_allowed
        self.force_L1_norm = force_L1_norm

    def generate_grid(self, constraint):
        # Generate lambda vectors for each event
        self.dim = len(constraint.basis["+"].columns)
        coefs = self._generate_coefs()
        # Convert the grid of basis coefficients into a grid of lambda vectors
        grid = constraint.basis["+"].dot(coefs["+"]) + constraint.basis["-"].dot(
            coefs["-"]
        )
        return grid

    def _get_true_dim(self):
        if self.force_L1_norm:
            true_dim = self.dim - 1
        else:
            true_dim = self.dim

        n_units = (float(self.grid_size) / (2.0 ** self.neg_allowed.sum())) ** (
            1.0 / true_dim
        ) - 1
        n_units = int(np.floor(n_units))
        if n_units < 0:
            n_units = 0

        return n_units

    def _build_grid(self):
        """Create an integer grid"""
        max_value = self._get_true_dim()
        self.accumulator = []
        while True:
            self.entry = np.zeros(self.dim)
            self._accumulate_integer_grid(0, max_value)
            max_value += 1
            if len(self.accumulator) >= self.grid_size:
                break

        xs = np.array(self.accumulator)
        xs = xs * self.grid_limit / max_value
        return xs

    def _accumulate_integer_grid(self, index, max_val):
        if index == self.dim:
            self.accumulator.append(self.entry.copy())
        else:
            if (index == self.dim - 1) and (self.force_L1_norm):
                if self.neg_allowed[index] and max_val > 0:
                    values = [-max_val, max_val]
                else:
                    values = [max_val]
            else:
                min_val = -max_val if self.neg_allowed[index] else 0
                values = range(min_val, max_val + 1)

            for current_value in values:
                self.entry[index] = current_value
                self._accumulate_integer_grid(index + 1, max_val - abs(current_value))

    def _generate_coefs(self):
        np_grid_values = self._build_grid()
        grid_values = pd.DataFrame(np_grid_values[: self.grid_size]).T
        pos_grid_values = grid_values.copy()
        neg_grid_values = -grid_values.copy()
        pos_grid_values[grid_values < 0] = 0
        neg_grid_values[grid_values < 0] = 0
        lambda_vector = {"+": pos_grid_values, "-": neg_grid_values}
        return lambda_vector



from typing import Optional

import numpy as np
import pandas as pd
import torch

from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

from .dataset import ADDataset
from .models import ADModel, ClassifierModel
from .trainer import TrainArgs, Trainer


class AdversarialDebiasing(BMImp):
    """
    Adversarial debiasing learns a classifier to maximize prediction accuracy and simultaneously reduce an
    adversary's ability to determine the protected attribute from the predictions. This approach leads to a fair classifier as the
    predictions cannot carry any group discrimination information that the adversary can exploit.

    Obs: Pytorch must be installed in order to use this techinique (pytorch = ">=1.12.1").

    References:
        B. H. Zhang, B. Lemoine, and M. Mitchell, "Mitigating Unwanted
        Biases with Adversarial Learning," AAAI/ACM Conference on Artificial
        Intelligence, Ethics, and Society, 2018.
    """

    def __init__(
        self,
        features_dim: int = None,
        keep_prob: Optional[float] = 0.1,
        hidden_size: Optional[int] = 128,
        batch_size: Optional[int] = 32,
        shuffle: Optional[bool] = True,
        epochs: Optional[int] = 10,
        initial_lr: Optional[float] = 0.01,
        use_debias: Optional[bool] = True,
        adversary_loss_weight: Optional[float] = 0.1,
        verbose: Optional[int] = 1,
        print_interval: Optional[int] = 100,
        device: Optional[str] = "cpu",
        seed: Optional[int] = None,
    ):

        """
        Parameters
        ----------
        features_dim: int
            Number of input feature X: (n_samples, features_dim)

        keep_prob: float
            Dropout parameter for classifier

        hidden_size: int
            Number of neurons on hidden layer

        batch_size: int
            Numer of examples used for each iteration

        shuffle: bool
            Shuffle data after each epoch

        epochs: int
            Number of epochs

        initial_lr: float
            Initial Learning Rate

        use_debias: bool
            If False Train a simple classifier

        adversary_loss_weight: float
            Adversarial Loss importance

        verbose : int
            Log progress if value > 0.

        print_interval : int
            Each `print_interval` steps print information.

        device: str
            pytorch paramter ("cpu", "cuda")

        seed: int
            seed for random state

        Return
        ------
        self
        """
        # default classifier config
        self.features_dim = features_dim
        self.keep_prob = keep_prob
        self.hidden_size = hidden_size

        # training config
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.epochs = epochs
        self.initial_lr = initial_lr

        # bias config
        self.adversary_loss_weight = adversary_loss_weight
        self.use_debias = use_debias

        # other configs
        self.verbose = verbose
        self.print_interval = print_interval
        self.device = device
        self.seed = seed

        self.sens_groups = SensitiveGroups()

    def transform_estimator(self, estimator=None):
        if estimator is None:
            self.estimator = ClassifierModel(
                self.features_dim, self.hidden_size, self.keep_prob
            )
        else:
            self.estimator = estimator
        return self

    def _create_dataset(self, X, y, sensitive_features):
        """
        Organize data for pytorch environment.
        """
        groups_num = np.array(
            self.sens_groups.fit_transform(sensitive_features, convert_numeric=True)
        )
        dataset = ADDataset(X, y, groups_num, device=self.device)
        return dataset

    def _build_trainer(self, dataset):
        """
        Create a model trainer.
        """

        adm = ADModel(estimator=self.estimator).to(self.device)

        train_args = TrainArgs(
            epochs=self.epochs,
            adversary_loss_weight=self.adversary_loss_weight,
            batch_size=self.batch_size,
            shuffle_data=self.shuffle,
            verbose=self.verbose,
            print_interval=self.print_interval,
        )

        trainer = Trainer(
            model=adm,
            dataset=dataset,
            train_args=train_args,
            use_debias=self.use_debias,
        )
        return trainer

    def fit(
        self,
        X: np.ndarray,
        y_true: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair classifier.

        Parameters
        ----------

        X : numpy array
            input matrix

        y_true : numpy array
            Target vector

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
        the same object
        """
        params = self._load_data(y_true=y_true, group_a=group_a, group_b=group_b)
        y_true = params["y_true"]
        group_a = params["group_a"]
        group_b = params["group_b"]
        self.classes_ = params["classes_"]
        sensitive_features = np.stack([group_a, group_b], axis=1)

        if self.seed is not None:
            torch.manual_seed(self.seed)

        dataset = self._create_dataset(X, y_true, sensitive_features)
        self.trainer = self._build_trainer(dataset)

        self.trainer.train()

        return self

    def predict(self, X):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted output per sample.
        """
        p = self.predict_proba(X)
        return np.argmax(p, axis=1).ravel()

    def predict_proba(self, X):
        """
        Prediction

        Description
        ----------
        Predict matrix probability for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted matrix probability per sample.
        """
        proba = np.empty((X.shape[0], 2))
        proba[:, 1] = self._forward(X)
        proba[:, 0] = 1.0 - proba[:, 1]
        return proba

    def predict_score(self, X):
        """
        Prediction

        Description
        ----------
        Predict probability for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted probability per sample.
        """
        p = self._forward(X).reshape([-1])
        return p

    def _preprocessing_data(self, X):
        X = np.array(X)
        X = torch.tensor(X).type(torch.FloatTensor)[None, :]
        return X

    def _forward(self, X):
        X = self._preprocessing_data(X)
        with torch.no_grad():
            p = self.estimator(X, trainable=False).ravel()
        return p



from statistics import mean

import numpy as np
import torch
import torch.optim as optim
from torch import nn
from torch.utils.data import DataLoader

from holisticai.bias.mitigation.inprocessing.commons import Logging


class TrainArgs:
    """
    training configuration for Trainer Class
    """

    def __init__(
        self,
        initial_epoch=0,
        epochs=10,
        adversary_loss_weight=0.1,
        initial_lr=0.001,
        verbose=0,
        print_interval=1,
        batch_size=32,
        shuffle_data=True,
    ):
        self.epochs = epochs
        self.adversary_loss_weight = adversary_loss_weight
        self.initial_lr = initial_lr
        self.initial_epoch = initial_epoch
        self.verbose = verbose
        self.print_interval = print_interval
        self.batch_size = batch_size
        self.shuffle_data = shuffle_data


class Trainer:
    """
    Trainer class support traditional classifier training and adverarial training.
    """

    def __init__(self, model, dataset, train_args, use_debias):
        self.train_args = train_args
        self.model = model
        self.use_debias = use_debias
        self.build(dataset)
        log_params = [
            ("iteration", int),
            ("classifier loss", float),
            ("adversarial loss", float),
        ]
        self.steps_per_epoch = (
            len(dataset) + self.train_args.batch_size - 1
        ) // self.train_args.batch_size
        self.total_iterations = self.steps_per_epoch * self.train_args.epochs
        self.logger = Logging(
            log_params=log_params,
            total_iterations=self.total_iterations,
            logger_format="epochs",
            epochs=self.train_args.epochs,
        )

    def build(self, dataset):
        """Build loss function and optimizers"""

        self.trainloader = DataLoader(
            dataset=dataset,
            batch_size=self.train_args.batch_size,
            shuffle=self.train_args.shuffle_data,
        )

        self.loss_cls_fn = nn.BCELoss()
        self.loss_adv_fn = nn.BCELoss()

        self.optimizer_cls = optim.SGD(
            self.model.classifier.parameters(),
            lr=self.train_args.initial_lr,
            momentum=0.9,
        )
        self.optimizer_adv = optim.SGD(
            self.model.adversarial.parameters(),
            lr=self.train_args.initial_lr,
            momentum=0.9,
        )

        self.scheduler_cls = torch.optim.lr_scheduler.LinearLR(
            self.optimizer_cls, start_factor=0.99, total_iters=20
        )
        self.scheduler_adv = torch.optim.lr_scheduler.LinearLR(
            self.optimizer_adv, start_factor=0.99, total_iters=20
        )

    def train(self):
        if self.use_debias:
            train_step = self._train_step_with_debias
        else:
            train_step = self._train_step_without_debias

        """Traditional one epoch"""
        for epoch in range(self.train_args.epochs):
            cls_loss, adv_loss = [], []
            for it, data in enumerate(self.trainloader, 1):

                loss_cls, loss_adv = train_step(data)

                cls_loss.append(loss_cls.item())
                adv_loss.append(0 if loss_adv is None else loss_adv.item())

                self._logging_progress(epoch, it, mean(cls_loss), mean(adv_loss))

            self.scheduler_cls.step()
            self.scheduler_adv.step()

    def _train_step_with_debias(self, data):
        """one step advesarial training classifier"""
        inputs, labels = data
        outputs = self.model(*inputs)

        # grad classifier Loss - grad adversarial Loss
        loss_adv = self.loss_adv_fn(outputs[1], labels[1])
        classifier_params = list(self.model.classifier.parameters())
        dloss_adv = torch.autograd.grad(
            outputs=loss_adv, inputs=classifier_params, retain_graph=True
        )

        loss_cls = self.loss_cls_fn(outputs[0], labels[0])
        dloss_cls = torch.autograd.grad(outputs=loss_cls, inputs=classifier_params)

        self._update_gradients(classifier_params, dloss_cls, dloss_adv)
        self.optimizer_cls.step()

        # grad adversarial Loss
        outputs = self.model(*inputs)
        loss_adv = self.loss_adv_fn(outputs[1], labels[1])
        adversarial_params = list(self.model.adversarial.parameters())
        dloss_adv = torch.autograd.grad(
            outputs=loss_adv, inputs=adversarial_params, retain_graph=True
        )
        for param, grad in zip(adversarial_params, dloss_adv):
            param.grad = grad
        self.optimizer_adv.step()

        return loss_cls, loss_adv

    def _train_step_without_debias(self, data):
        """Traditional classifier tran step"""
        inputs, labels = data
        outputs = self.model(*inputs)

        # Classifier
        loss_cls = self.loss_cls_fn(outputs[0], labels[0])
        classifier_params = list(self.model.classifier.parameters())
        dloss_cls = torch.autograd.grad(outputs=loss_cls, inputs=classifier_params)
        for param, grad in zip(classifier_params, dloss_cls):
            param.grad = grad
        self.optimizer_cls.step()

        return loss_cls, None

    def _update_gradients(self, classifier_params, dloss_cls, dloss_adv):
        """update classifier gradients with adversarial model"""
        normalize = lambda x: x / (torch.norm(x) + np.finfo(np.float32).tiny)
        for param, grad, grad_adv in zip(classifier_params, dloss_cls, dloss_adv):
            unit_adversary_grad = normalize(grad_adv)
            grad -= torch.sum(grad * unit_adversary_grad) * unit_adversary_grad
            grad -= self.train_args.adversary_loss_weight * grad_adv
            param.grad = grad

    def _logging_progress(self, epoch, it, cls_loss, adv_loss):
        iteration = epoch * self.steps_per_epoch + it
        if self.train_args.verbose > 0:
            if (iteration % self.train_args.print_interval) == 0 or (
                iteration % self.total_iterations
            ) == 0:
                self.logger.update(iteration, cls_loss, adv_loss)



import math
import random
from copy import deepcopy
from typing import Union

import numpy as np

from ._categorical_feature import CategoricalFeature
from ._sparse_list import SparseList


def get_categories_count_norm(
    categories, all_stratified_groups, count_dict, group_features
):
    """
    Description
    -----------
    Find the normalized count for each category, where normalized count is count divided by
    the number of people in that group.
    """
    norm = {
        cat: SparseList(
            data=(
                count_dict[cat][i] * (1.0 / len(group_features[group].data))
                if group_features[group].data
                else 0.0
                for i, group in enumerate(all_stratified_groups)
            )
        )
        for cat in categories
    }
    return norm


def gen_desired_dist(
    group_index, cat, col_id, median, repair_level, norm_counts, feature_to_remove, mode
):
    """
    Description
    -----------
    Generate the desired distribution and desired "count" for a given group-category-feature combination.
    """
    if feature_to_remove == col_id:
        return 1 if cat == mode else (1 - repair_level) * norm_counts[cat][group_index]
    else:
        return (1 - repair_level) * norm_counts[cat][group_index] + (
            repair_level * median[cat]
        )


def assign_overflow(
    all_stratified_groups, categories, overflow, group_features, repair_generator
):
    """
    Description
    -----------
    Assign overflow observations to categories based on the group's desired distribution.
    """
    feature = deepcopy(group_features)
    assigned_overflow = {}
    desired_dict_list = {}
    for group_index, group in enumerate(all_stratified_groups):
        # Calculate the category proportions.
        dist_generator = lambda cat: repair_generator(group_index, cat)
        cat_props = list(map(dist_generator, categories))

        if all(elem == 0 for elem in cat_props):  # TODO: Check that this is correct!
            cat_props = [1.0 / len(cat_props)] * len(cat_props)
        s = float(sum(cat_props))
        cat_props = [elem / s for elem in cat_props]
        desired_dict_list[group] = cat_props
        assigned_overflow[group] = {}
        for i in range(int(overflow[group])):
            distribution_list = desired_dict_list[group]
            number = random.uniform(0, 1)
            cat_index = 0
            tally = 0
            for j in range(len(distribution_list)):
                value = distribution_list[j]
                if number < (tally + value):
                    cat_index = j
                    break
                tally += value
            assigned_overflow[group][i] = categories[cat_index]
        # Actually do the assignment
        count = 0
        for i, value in enumerate(group_features[group].data):
            if value == 0:
                (feature[group].data)[i] = assigned_overflow[group][count]
                count += 1
    return feature, assigned_overflow, desired_dict_list


def get_categories_count(categories, all_stratified_groups, group_feature):
    """
    Description
    -----------
    Count the observations in each category.
    """
    count_dict = {
        cat: SparseList(
            data=(
                group_feature[group].category_count[cat]
                if cat in group_feature[group].category_count
                else 0
                for group in all_stratified_groups
            )
        )
        for cat in categories
    }

    return count_dict


def gen_desired_count(
    group_index, group, category, median, group_features, repair_level, categories_count
):
    med = median[category]
    size = len(group_features[group].data)
    count = categories_count[category][group_index]
    des_count = math.floor(((1 - repair_level) * count) + (repair_level) * med * size)
    return des_count


def flow_on_group_features(all_stratified_groups, group_features, repair_generator):
    """
    Description
    -----------
    Run Max-flow to distribute as many observations to categories as possible.
    Overflow are those observations that are left over
    """
    dict1 = {}
    dict2 = {}
    for i, group in enumerate(all_stratified_groups):
        feature = group_features[group]
        count_generator = lambda category: repair_generator(i, group, category)
        DG = feature.create_graph(count_generator)
        new_feature, overflow = feature.repair(DG)
        dict2[group] = overflow
        dict1[group] = new_feature

    return dict1, dict2


def get_count(cat, group_feature_category_count):
    if cat in group_feature_category_count:
        return group_feature_category_count[cat]
    else:
        return 0


def get_count_norm(count, group_feature_data):
    if group_feature_data:
        return count * (1.0 / len(group_feature_data))
    else:
        return 0.0


def get_column_type(values: Union[list, np.ndarray]):
    if all(isinstance(value, float) for value in values):
        return float
    elif all(isinstance(value, int) for value in values):
        return int
    else:
        return str


def get_median(values, kdd):

    if not values:
        raise Exception("Cannot calculate median of list with no values!")

    sorted_values = deepcopy(values)
    sorted_values.sort()  # Not calling `sorted` b/c `sorted_values` may not be list.

    if kdd:
        return sorted_values[len(values) // 2]
    else:
        if len(values) % 2 == 0:
            return sorted_values[len(values) // 2 - 1]
        else:
            return sorted_values[len(values) // 2]


def get_group_data(all_stratified_groups, stratified_group_data, col_id):
    group_features = {}
    for group in all_stratified_groups:
        points = [
            (i, val)
            for indices, val in stratified_group_data[group][col_id]
            for i in indices
        ]
        points = sorted(points, key=lambda x: x[0])  # Sort by index
        values = [value for _, value in points]

        # send values to CategoricalFeature object, which bins the data into categories
        group_features[group] = CategoricalFeature(values)
    return group_features


# Find the median normalized count for each category
def get_median_per_category(categories, categories_count_norm):
    return {cat: get_median(categories_count_norm[cat], False) for cat in categories}


def get_mode(values):
    counts = {}
    for value in values:
        counts[value] = 1 if value not in counts else counts[value] + 1
    mode_tuple = max(list(counts.items()), key=lambda tup: tup[1])
    return mode_tuple[0]


def make_histogram_bins(bin_size_calculator, data, col_id):
    feature_vals = [row[col_id] for row in data]
    bin_range = bin_size_calculator(feature_vals)

    if bin_range == 0.0:
        bin_range = 1.0

    data_tuples = list(enumerate(data))  # [(0,row), (1,row'), (2,row''), ... ]
    sorted_data_tuples = sorted(data_tuples, key=lambda tup: tup[1][col_id])

    max_val = max(data, key=lambda datum: datum[col_id])[col_id]
    min_val = min(data, key=lambda datum: datum[col_id])[col_id]

    index_bins = []
    val_ranges = []
    curr = min_val
    while curr <= max_val:
        index_bins.append([])
        val_ranges.append((curr, curr + bin_range))
        curr += bin_range

    for row_index, row in sorted_data_tuples:
        for bin_num, val_range in enumerate(val_ranges):
            if val_range[0] <= row[col_id] < val_range[1]:
                index_bins[bin_num].append(row_index)
                break

    index_bins = [b for b in index_bins if b]

    return index_bins


def FreedmanDiaconisBinSize(feature_values):
    q75, q25 = np.percentile(feature_values, [75, 25])
    IQR = q75 - q25
    return 2.0 * IQR * len(feature_values) ** (-1.0 / 3.0)



import random
from typing import Optional

import pandas as pd


def generate_rankings(
    M: Optional[int], k: Optional[int], p: Optional[float], return_p_attr=False
):
    """
    Generates M rankings of k elements (candidates) using Yang-Stoyanovich process

    Parameters
    ----------
    M: int
        Number of rankings to generate
    k: int
        Number of elements should each ranking have
    p: float
        Probability that a candidate is protected

    Return
    ------
        DataFrame or Tuple of DataFrames
    """
    rankings = []
    for m in range(M):
        for i in range(k):
            is_protected = random.random() <= p
            rankings.append(
                {"X": m, "Y": k - i, "score": k - i, "protected": is_protected}
            )

    df = pd.DataFrame(rankings)
    if return_p_attr:
        return df[["X", "Y", "score"]], df[["X", "Y", "protected"]]
    return df[["X", "Y", "score", "protected"]]



from typing import Optional

import numpy as np
import pandas as pd

from holisticai.utils.transformers.bias import BMInprocessing as BMImp

from .algorithm import FairRecAlg


class FairRec(BMImp):
    """
    FairRecommendationSystem (FairRec), exhibes the desired two-sided fairness by
    mapping the fair recommendation problem to a fair allocation problem; moreover,
    it is agnostic to the specifics of the data-driven model (that estimates the
    product-customer relevance scores) which makes it more scalable and easy to adapt [1].
    References:
        [1] Patro, Gourab K., et al. "Fairrec: Two-sided fairness for personalized
        recommendations in two-sided platforms." Proceedings of The Web Conference 2020. 2020.
    """

    def __init__(
        self, rec_size: Optional[int] = 10, MMS_fraction: Optional[float] = 0.5
    ):
        """
        Init FairRec algorithm
        Parameters
        ----------
        rec_size : int
            Specifies the number of recommended items.
        MMS_fraction : float
            Maximin Share (MMS) threshold of producers exposure.
        """

        self.rec_size = rec_size
        self.MMS_fraction = MMS_fraction

    def fit(self, X):
        algorithm = FairRecAlg(rec_size=self.rec_size, MMS_fraction=self.MMS_fraction)
        self.recommendation = algorithm.rank(X)
        return self

    def predict(self, X: Optional[np.ndarray], top_n: Optional[int] = None):
        """
        Fit model

        Parameters
        ----------
        X : matrix-like
            scored matrix, 0 means non-raked cases.

        Returns
        -------
        recommendations : dict
            A dictionary of recommendations for each user.
        """
        if top_n is None:
            algorithm = FairRecAlg(
                rec_size=self.rec_size, MMS_fraction=self.MMS_fraction
            )
            self.recommendation = algorithm.rank(X)

        dfs = []
        for i, key in enumerate(self.recommendation.keys()):
            df = pd.DataFrame()
            df["Y"] = np.array(self.recommendation[key])
            df["X"] = i
            df["score"] = 1.0
            dfs.append(df)
        dfs = pd.concat(dfs, axis=0)
        return dfs[["X", "Y", "score"]]



"""
The :mod:`holisticai.datasets` module includes dataloaders for quick experimentation
"""

from ._dataloaders import (
    load_adult,
    load_last_fm,
    load_law_school,
    load_student,
    load_us_crime,
)

__all__ = [
    "load_adult",
    "load_student",
    "load_law_school",
    "load_last_fm",
    "load_us_crime",
]



import numpy as np

from .utils import calculate_popularity_model


def constant_propensity(rmat, numItems):
    numObservations = np.ma.count(rmat)
    numUsers, numItems = np.shape(rmat)
    scale = numUsers * numItems
    inversePropensities = (
        np.ones((numUsers, numItems), dtype=np.longdouble) * scale / numObservations
    )
    return inversePropensities


def popularity_model_propensity(rmat):
    prop_score = calculate_popularity_model(rmat)
    IversPropensity = [prop_score] * rmat.shape[0]
    return np.array(IversPropensity)[:, 0, :]



import cvxpy as cp
import numpy as np

from .utils import *


class FairScoreClassifierAlgorithm:
    """
    Generates a classification model that integrates fairness constraints for multiclass classification. This algorithm
    returns a matrix of lambda coefficients that scores a given input vector. The higher the score, the higher the probability
    of the input vector to be classified as the majority class.

    References:
        Julien Rouzot, Julien Ferry, Marie-José Huguet. Learning Optimal Fair Scoring Systems for Multi-
        Class Classification. ICTAI 2022 - The 34th IEEE International Conference on Tools with Artificial
        Intelligence, Oct 2022, Virtual, United States.
    """

    def __init__(
        self,
        objectives: dict,
        fairness_groups: list,
        fairness_labels: list,
        constraints: dict = {},
        lambda_bound: int = 9,
        time_limit: int = 100,
    ):
        """
        Init FairScoreClassifier object

        Parameters
        ----------
        objectives : dict
            The weighted objectives list to be optimized.

        fairness_groups : list
            The sensitive groups indexes.

        fairness_labels : list
            The senstive labels indexes.

        constraints : dict
            The constraints list to be used in the optimization. The keys are the constraints names and the values are the bounds.

        lambda_bound : int
            Lower and upper bound for the scoring system cofficients.
        """
        self.objectives = objectives
        self.fairness_groups = fairness_groups
        self.fairness_labels = fairness_labels
        self.constraints = constraints
        self.lambda_bound = lambda_bound
        self.time_limit = time_limit

    def fit(self, X, y):
        print("Start fitting")

        N_class = get_class_count(y)
        class_indexes = get_class_indexes(y)

        self.lambdas = self.solve_model(X, y, N_class, class_indexes)

        accuracy = get_accuracy(X, y, self.lambdas)
        balanced_accuracy = get_balanced_accuracy(X, y, self.lambdas)
        print(f"Accuracy : {accuracy}")
        print(f"Balanced accuracy : {balanced_accuracy}")

    def solve_model(self, X, y, N_class, class_indexes):
        N = len(X)
        L = len(y[0])
        D = len(X[0])
        gamma = 0.01
        M = self.lambda_bound * D + 1

        l = cp.Variable((L, D))
        constraints = [l <= self.lambda_bound, l >= -self.lambda_bound]

        for g in self.fairness_groups:
            constraints.append(l[:, g] == 0)

        z = cp.Variable(N, boolean=True)

        if "s" in self.constraints:
            alpha = cp.Variable((L, D), boolean=True)

        if (
            "omr" in self.constraints
            or "sp" in self.constraints
            or "pe" in self.constraints
            or "eo" in self.constraints
            or "eod" in self.constraints
        ):
            pos = cp.Variable((N, L), boolean=True)

        y_idx = np.argmax(y, axis=1)
        if "a" in self.objectives or "ba" in self.objectives:
            for i in range(N):
                for offset in range(1, L):
                    constraints.append(
                        -M * z[i]
                        <= cp.sum(l[y_idx[i], :] @ X[i, :])
                        - gamma * y_idx[i]
                        - cp.sum(l[(y_idx[i] + offset) % L, :] @ X[i, :])
                        - gamma * ((y_idx[i] + offset) % L)
                    )

        if "s" in self.constraints:
            constraints.append(-self.lambda_bound * alpha <= l)
            constraints.append(l <= self.lambda_bound * alpha)

        if (
            "omr" in self.constraints
            or "sp" in self.constraints
            or "pe" in self.constraints
            or "eo" in self.constraints
            or "eod" in self.constraints
        ):
            for i in range(N):
                for index in range(L):
                    for offset in range(1, L):
                        constraints.append(
                            -M * (1 - pos[i, index])
                            <= cp.sum(l[index, :] @ X[i, :])
                            - gamma * index
                            - cp.sum(l[(index + offset) % L, :] @ X[i, :])
                            - gamma * ((index + offset) % L)
                        )
                constraints.append(cp.sum(pos[i, :]) == 1)

        if "s" in self.constraints:
            for index in range(L):
                constraints.append(cp.sum(alpha[index, :]) <= self.constraints["s"])

        for index in self.fairness_labels:
            for g in self.fairness_groups:

                i_g = [i for i in range(N) if X[i][g] == 1]
                i_g_bar = [i for i in range(N) if X[i][g] == 0]
                N_g = len(i_g)
                N_g_bar = len(i_g_bar)

                i_g_pos = [i for i in range(N) if X[i][g] == 1 and y[i][index] == 1]
                i_g_neg = [i for i in range(N) if X[i][g] == 1 and y[i][index] == 0]
                N_g_pos = len(i_g_pos)
                N_g_neg = len(i_g_neg)

                i_g_bar_pos = [i for i in range(N) if X[i][g] == 0 and y[i][index] == 1]
                i_g_bar_neg = [i for i in range(N) if X[i][g] == 0 and y[i][index] == 0]
                N_g_bar_pos = len(i_g_bar_pos)
                N_g_bar_neg = len(i_g_bar_neg)

                if N_g == 0 or N_g_bar == 0:
                    print(
                        "At least one of the protected groups is empty, skipping fairness constraints"
                    )
                    break

                if "omr" in self.constraints:
                    constraints.append(
                        -self.constraints["omr"]
                        <= (1 / N_g) * cp.sum(pos[i_g_neg, index])
                        + cp.sum(1 - pos[i_g_pos, index])
                        - (1 / N_g_bar) * cp.sum(pos[i_g_bar_neg, index])
                        + cp.sum(1 - pos[i_g_bar_pos, index])
                    )
                    constraints.append(
                        (1 / N_g) * cp.sum(pos[i_g_neg, index])
                        + cp.sum(1 - pos[i_g_pos, index])
                        - (1 / N_g_bar) * cp.sum(pos[i_g_bar_neg, index])
                        + cp.sum(1 - pos[i_g_bar_pos, index])
                        <= self.constraints["omr"]
                    )

                if "sp" in self.constraints and i_g and i_g_bar:
                    constraints.append(
                        -self.constraints["sp"]
                        <= (1 / N_g) * cp.sum(pos[i_g, index])
                        - (1 / N_g_bar) * cp.sum(pos[i_g_bar, index])
                    )
                    constraints.append(
                        (1 / N_g) * cp.sum(pos[i_g, index])
                        - (1 / N_g_bar) * cp.sum(pos[i_g_bar, index])
                        <= self.constraints["sp"]
                    )

                if "pe" in self.constraints and i_g_neg and i_g_bar_neg:
                    constraints.append(
                        -self.constraints["pe"]
                        <= (1 / N_g_neg) * cp.sum(pos[i_g_neg, index])
                        - (1 / N_g_bar_neg) * cp.sum(pos[i_g_bar_neg, index])
                    )
                    constraints.append(
                        (1 / N_g_neg) * cp.sum(pos[i_g_neg, index])
                        - (1 / N_g_bar_neg) * cp.sum(pos[i_g_bar_neg, index])
                        <= self.constraints["pe"]
                    )

                if "eo" in self.constraints and i_g_pos and i_g_bar_pos:
                    constraints.append(
                        -self.constraints["eo"]
                        <= (1 / N_g_pos) * cp.sum(pos[i_g_pos, index])
                        - (1 / N_g_bar_pos) * cp.sum(pos[i_g_bar_pos, index])
                    )
                    constraints.append(
                        (1 / N_g_pos) * cp.sum(pos[i_g_pos, index])
                        - (1 / N_g_bar_pos) * cp.sum(pos[i_g_bar_pos, index])
                        <= self.constraints["eo"]
                    )

                if (
                    "eod" in self.constraints
                    and i_g_neg
                    and i_g_bar_neg
                    and i_g_pos
                    and i_g_bar_pos
                ):
                    constraints.append(
                        -self.constraints["eod"]
                        <= (1 / N_g_neg) * cp.sum(pos[i_g_neg, index])
                        - (1 / N_g_bar_neg) * cp.sum(pos[i_g_bar_neg, index])
                    )
                    constraints.append(
                        (1 / N_g_neg) * cp.sum(pos[i_g_neg, index])
                        - (1 / N_g_bar_neg) * cp.sum(pos[i_g_bar_neg, index])
                        <= self.constraints["eod"]
                    )
                    constraints.append(
                        -self.constraints["eod"]
                        <= (1 / N_g_pos) * cp.sum(pos[i_g_pos, index])
                        - (1 / N_g_bar_pos) * cp.sum(pos[i_g_bar_pos, index])
                    )
                    constraints.append(
                        (1 / N_g_pos) * cp.sum(pos[i_g_pos, index])
                        - (1 / N_g_bar_pos) * cp.sum(pos[i_g_bar_pos, index])
                        <= self.constraints["eod"]
                    )

        if "a" in self.objectives and not ("s" in self.constraints):
            cost = (1 / N) * cp.sum(z)

        if "ba" in self.objectives and not ("s" in self.constraints):
            cost = 0
            for i, indexes in enumerate(class_indexes):
                cost += cp.sum(z[indexes]) / N_class[i]
            cost /= len(N_class)

        if "a" in self.objectives and "s" in self.constraints:
            cost = (1 / N) * cp.sum(z) + (1 / (self.constraints["s"] * L * N)) * cp.sum(
                alpha
            )

        if "ba" in self.objectives and "s" in self.constraints:
            cost = 0
            for i, indexes in enumerate(class_indexes):
                cost += cp.sum(z[indexes]) / N_class[i]
            cost /= len(N_class)
            cost += (1 / (self.constraints["s"] * L * N)) * cp.sum(alpha)

        prob = cp.Problem(cp.Minimize(cost), constraints)
        prob.solve(solver=cp.CBC, verbose=True, maximumSeconds=self.time_limit)
        return l.value

    def predict(self, X):
        """
        Returns the predictions of the set of scoring systems for the given entries

        X : The input features matrix
        """

        y = []

        for index, sample in enumerate(X):
            scores = []
            for l_list in self.lambdas:
                scores.append(
                    sum(feature * l_list[j] for j, feature in enumerate(sample))
                )
            y_pred = []
            for i in range(len(scores)):
                if i == np.argmax(scores):
                    y_pred.append(1)
                else:
                    y_pred.append(0)
            y.append(y_pred)

        return np.array(y)



import numpy as np
from sklearn.base import TransformerMixin

from holisticai.utils.transformers.bias import BMPreprocessing as BMPre


class CorrelationRemover(BMPre):
    """
    CorrelationRemover applies a linear transformation to the non-sensitive feature columns
    in order to remove their correlation with the sensitive feature columns while retaining
    as much information as possible (as measured by the least-squares error).

    Notes
    -----
    This method will change the original dataset by removing all correlation with sensitive
    values. Note that the lack of correlation does not imply anything about statistical dependence.
    Therefore, it is expected this to be most appropriate as a preprocessing step for
    (generalized) linear models.

    Reference
    ---------

    """

    def __init__(self, alpha=1):
        """
        Parameters
        ----------
            alpha : float
                parameter to control how much to filter, for alpha=1.0 we filter out
                all information while for alpha=0.0 we don't apply any.
        """
        self.alpha = alpha

    def fit(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Fit.

        Description
        -----------
        Learn the projection required to make the dataset uncorrelated with groups (group_a and group_b).

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Return
        ------
            Self
        """
        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_features = np.stack([group_a, group_b], axis=1).astype(np.int32)
        self.sensitive_mean_ = sensitive_features.mean()
        sensitive_features_center = sensitive_features - self.sensitive_mean_
        self.beta_, _, _, _ = np.linalg.lstsq(sensitive_features_center, X, rcond=None)
        self.X_shape_ = X.shape
        return self

    def transform(self, X: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):

        """
        Description
        ----------
        Transform X by applying the correlation remover.

        Parameters
        ----------
        X : matrix-like
            Input matrix
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
            np.ndarray
        """

        params = self._load_data(X=X, group_a=group_a, group_b=group_b)
        X = params["X"]
        group_a = params["group_a"]
        group_b = params["group_b"]

        sensitive_features = np.stack([group_a, group_b], axis=1).astype(np.int32)
        self.sensitive_mean_ = sensitive_features.mean()
        sensitive_features_center = sensitive_features - self.sensitive_mean_
        X_filtered = X - sensitive_features_center.dot(self.beta_)
        X = np.atleast_2d(X)
        X_filtered = np.atleast_2d(X_filtered)
        return self.alpha * X_filtered + (1 - self.alpha) * X

    def fit_transform(
        self,
        X: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        X : matrix-like
            Input data
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Return
        ------
            Self
        """
        return self.fit(X, group_a, group_b).transform(X, group_a, group_b)



from typing import Optional

import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost

from .algorithm import PluginEstimationAndCalibrationAlgorithm


class PluginEstimationAndCalibration(BMPost):
    """
    Plugin Estimation and Calibration postprocessing optimizes over calibrated regressor outputs via a
    smooth optimization. The rates of convergence of the proposed estimator were derived in terms of
    the risk and fairness constraint.

    References:
        Chzhen, Evgenii, et al. "Fair regression via plug-in estimator and recalibration with statistical
        guarantees." Advances in Neural Information Processing Systems 33 (2020): 19137-19148.
    """

    def __init__(self, L: Optional[int] = 25, beta: Optional[float] = 0.1):
        """Create a Calibrated Equalized Odds Post-processing instance."""
        self.algorithm = PluginEstimationAndCalibrationAlgorithm(L=L, beta=beta)

    def fit(self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Compute parameters for calibrated equalized odds.

        Description
        ----------
        Compute parameters for calibrated equalized odds algorithm.

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (num_examples, ).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        Self
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        self.algorithm.fit(y_pred, sensitive_features)
        return self

    def transform(
        self,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use a fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        threshold : float
            float value to discriminate between 0 and 1

        Returns
        -------
        dictionnary with new predictions
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        sensitive_features = np.stack([group_a, group_b], axis=1)

        new_y_pred = self.algorithm.transform(y_pred, sensitive_features)

        return {"y_pred": new_y_pred}

    def fit_transform(
        self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_pred,
            group_a,
            group_b,
        ).transform(y_pred, group_a, group_b)



import numpy as np

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost

from .algorithm import WassersteinBarycenterAlgorithm


class WassersteinBarycenter(BMPost):
    """
    Fair Regression with Wasserstein Barycenters learning a real-valued function that
    satisfies the Demographic Parity constraint. The strategy founds the optimal fair
    predictor computing the Wasserstein barycenter of the distributions induced by the
    standard regression function on the sensitive groups.

    References:
        Chzhen, Evgenii, et al. "Fair regression with wasserstein barycenters."
        Advances in Neural Information Processing Systems 33 (2020): 7321-7331.
    """

    def __init__(self):
        """Create a Calibrated Equalized Odds Post-processing instance."""
        self.algorithm = WassersteinBarycenterAlgorithm()

    def fit(self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray):
        """
        Compute parameters for calibrated equalized odds.

        Description
        ----------
        Compute parameters for calibrated equalized odds algorithm.

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        Self
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]

        sensitive_features = np.stack([group_a, group_b], axis=1)
        self.algorithm.fit(y_pred, sensitive_features)
        return self

    def transform(
        self,
        y_pred: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use a fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        threshold : float
            float value to discriminate between 0 and 1

        Returns
        -------
        dictionnary with new predictions
        """
        params = self._load_data(y_pred=y_pred, group_a=group_a, group_b=group_b)

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        sensitive_features = np.stack([group_a, group_b], axis=1)

        new_y_pred = self.algorithm.transform(y_pred, sensitive_features)

        return {"y_pred": new_y_pred}

    def fit_transform(
        self, y_pred: np.ndarray, group_a: np.ndarray, group_b: np.ndarray
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (num_examples,).
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_pred,
            group_a,
            group_b,
        ).transform(y_pred, group_a, group_b)



import numpy as np
import pandas as pd

# sklearn imports
from sklearn.metrics import (
    adjusted_mutual_info_score,
    mean_absolute_error,
    silhouette_samples,
)

# scipy imports
from ...utils._recommender_tools import entropy

# utils
from ...utils._validation import _clustering_checks


def cluster_balance(group_a, group_b, y_pred):
    """
    Cluster Balance.

    Description
    -----------
    Given a clustering and protected attribute. The cluster balance is
    the minimum over all groups and clusters of the ratio of the representation
    of members of that group in that cluster to the representation overall.

    Interpretation
    --------------
    A value of 1 is desired. That is when all clusters have the exact same
    representation as the data. Lower values imply the existence of clusters
    where either group_a or group_b is underrepresented.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Cluster Balance

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cluster_balance
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> cluster_balance(group_a, group_b, y_pred_cluster)
    0.5
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _, _ = _clustering_checks(group_a, group_b, y_pred)

    # Get clusters
    clusters = np.unique(y_pred)

    # group_a ratio overall
    r_tot_a = group_a.sum() / len(group_a)
    # group_b ratio overall
    r_tot_b = group_b.sum() / len(group_b)

    min_ratio = 1

    # loop over clusters
    for c in clusters:
        # variables
        members = y_pred == c
        n_members = members.sum() + 1.0e-20
        n_a = group_a[members].sum() + 1.0e-20
        n_b = group_b[members].sum() + 1.0e-20

        # group_a ratios
        ratio_a = (n_a / n_members) / r_tot_a
        min_a = min(ratio_a, 1 / ratio_a)

        # group_b ratios
        ratio_b = (n_b / n_members) / r_tot_b
        min_b = min(ratio_b, 1 / ratio_b)

        min_ratio = min(min_ratio, min_a, min_b)

    # return minimum balance in list
    return min_ratio


def min_cluster_ratio(group_a, group_b, y_pred):
    """
    Minimum Cluster Ratio.

    Description
    -----------
    Given a clustering and protected attributes. The min cluster ratio is
    the minimum over all clusters of the ratio of number of group_a members
    to the number of group_b members.

    Interpretation
    --------------
    A value of 1 is desired. That is when all clusters are perfectly
    balanced. Low values imply the existence of clusters where
    group_a has fewer members than group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Minimum Cluster Ratio

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import min_cluster_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> min_cluster_ratio(group_a, group_b, y_pred_cluster)
    0.2499999999375
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _, _ = _clustering_checks(group_a, group_b, y_pred)

    # Get clusters
    clusters = np.unique(y_pred)
    min_ratio = np.Inf

    # Get balance of each cluster
    for c in clusters:
        members = y_pred == c
        n_a = group_a[members].sum()
        n_b = group_b[members].sum()

        min_ratio = min(min_ratio, (n_a / (n_b + 1.0e-32)))

    return min_ratio


def _avg_cluster_ratio(group_a, group_b, y_pred):
    """
    Average Cluster Ratio.

    Description
    -----------
    Given a clustering and protected attributes. The average cluster ratio is
    the average over all clusters of the ratio of group_a members to group_b
    members in that cluster.

    Interpretation
    --------------
    A value of 1 is desired. Low values imply the predominance of clusters where
    group_a is underrepresented. High values imply the predominance of clusters where
    group_b is underrepresented.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Average Cluster Ratio

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import avg_cluster_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> avg_cluster_ratio(group_a, group_b, y_pred_cluster)
    1.0833333323124998
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _, _ = _clustering_checks(group_a, group_b, y_pred)

    # Get clusters
    clusters = np.unique(y_pred)
    n_clusters = len(clusters)
    balances = np.zeros((n_clusters,))

    # Get balance of each cluster
    for i, c in zip(range(n_clusters), clusters):
        members = y_pred == c
        n_a = group_a[members].sum()
        n_b = group_b[members].sum()
        # if n_b zero we get infinity for mean
        if n_b == 0:
            return np.Inf
        balances[i] = n_a / n_b
        # Check non zero for b
        balances[i] = n_a / (n_b + 1.0e-32)

    return balances.mean()


def _cluster_dist(y_pred_g, clusters):
    """
    Group distribution over clusters.

    Description
    -----------
    This function computes the distribution of the group across clusters.

    Parameters
    ----------
    y_pred_g :
        Cluster predictions (categorical)
    clusters :
        Cluster ground truth (categorical)

    Returns
    -------
    numpy array
        Cluster Distribution
    """
    bin_mat = y_pred_g.reshape(-1, 1) == clusters.reshape(1, -1)
    dist = bin_mat.sum(axis=0)
    dist = dist / dist.sum()

    return dist


def cluster_dist_l1(group_a, group_b, y_pred):
    """
    Cluster Distribution Total Variation.

    Description
    -----------
    This function computes the distribution of group_a and group_b across clusters.
    It then outputs the total variation distance between these distributions.

    Interpretation
    --------------
    A value of 0 is desired. That indicates that both groups are distributed
    similarly amongst the clusters. The metric ranges between 0 and 1,
    with higher values indicating the groups are distributed in very
    different ways.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Cluster Distribution Total Variation

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cluster_dist_l1
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> cluster_dist_l1(group_a, group_b, y_pred_cluster)
    0.4166666666666667
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _, _ = _clustering_checks(group_a, group_b, y_pred)

    # get unique clusters
    clusters = np.unique(y_pred)

    # split data by group
    y_pred_a = y_pred[group_a == 1]
    y_pred_b = y_pred[group_b == 1]

    # compute distributions
    dist_a = _cluster_dist(y_pred_a, clusters)
    dist_b = _cluster_dist(y_pred_b, clusters)

    # return total variation norm
    return 0.5 * len(dist_a) * mean_absolute_error(dist_a, dist_b)


def cluster_dist_kl(group_a, group_b, y_pred):
    """
    Cluster Distribution KL.

    Description
    -----------
    This function computes the distribution of group_a and group_b
    membership across the clusters. It then returns the KL distance
    from the distribution of group_a to the distribution of group_b.

    Interpretation
    --------------
    A value of 0 is desired. That indicates that both groups are distributed
    similarly amongst the clusters. Higher values indicate the distributions
    of both groups amongst the clusters differ more.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Cluster Distribution KL : :math:`KL(P_a,P_b)`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cluster_dist_kl
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> cluster_dist_kl(group_a, group_b, y_pred_cluster)
    0.4054651081081642
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, _, _ = _clustering_checks(group_a, group_b, y_pred)

    # get unique clusters
    clusters = np.unique(y_pred)

    # split data by group
    y_pred_a = y_pred[group_a == 1]
    y_pred_b = y_pred[group_b == 1]

    # compute distributions
    dist_a = _cluster_dist(y_pred_a, clusters)
    dist_b = _cluster_dist(y_pred_b, clusters)

    # return KL
    return entropy(dist_a, dist_b)


def cluster_dist_entropy(group, y_pred):
    """
    Minority Cluster Distribution Entropy.

    Description
    -----------
    The entropy of the distribution of the group
    over the clusters.

    Interpretation
    --------------
    Lower values indicate most members of the group are allocated to
    the same cluaster. Hence we encourage higher values of
    the entropy, which indicate more homogeneity.

    Parameters
    ----------
    group : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)

    Returns
    -------
    float
        Group Presence Entropy : :math:`Entropy(P_{group})`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cluster_dist_entropy
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred_cluster = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> cluster_dist_entropy(group_b, y_pred_cluster)
    0.8675632284814613
    """
    # check and coerce inputs
    group, _, y_pred, _, _, _ = _clustering_checks(group, group, y_pred)

    # get unique clusters
    clusters = np.unique(y_pred)

    # split data by group
    y_pred_group = y_pred[group == 1]

    # compute distribution
    dist_b = _cluster_dist(y_pred_group, clusters)

    # return Entropy
    return entropy(dist_b)


def _ami_diff(group_a, group_b, y_pred, y_true):
    """
    Adjusted Mutual information Difference.

    Description
    -----------
    We compute the difference of the adjusted mutual information
    on group_a and group_b.

    Interpretation
    --------------
    The MI difference ranges from -1 to 1, with lower values indicating bias
    towards group_a and larger values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)
    y_true : array-like
        Cluster ground truth (categorical)

    Returns
    -------
    float
        Mutual information Difference

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import ami_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> y_true = np.array([0, 1, 1, 2, 1, 0, 1, 2, 0, 2])
    >>> ami_diff(group_a, group_b, y_pred, y_true)
    0.6342556627317533
    """
    # check and coerce inputs
    group_a, group_b, y_pred, y_true, _, _ = _clustering_checks(
        group_a, group_b, y_pred, y_true
    )

    # Slice by min and maj groups
    y_pred_a = y_pred[group_a == 1]
    y_pred_b = y_pred[group_b == 1]
    y_true_a = y_true[group_a == 1]
    y_true_b = y_true[group_b == 1]

    # Compute AMI scores
    ami_a = adjusted_mutual_info_score(y_true_a, y_pred_a)
    ami_b = adjusted_mutual_info_score(y_true_b, y_pred_b)

    # Return Spread
    return ami_a - ami_b


def social_fairness_ratio(group_a, group_b, data, centroids):
    """
    Social Fairness Ratio.

    Description
    -----------
    Given a centroid based clustering, this function compute the average
    distance to the nearest centroid for both groups. The metric is the
    ratio of the resulting distance for group_a to group_b.

    Interpretation
    --------------
    A value of 1 is desired. Lower values indicate the group_a
    is on average closer to the respective centroids. Higher
    values indicate that group_a is on average further from the
    respective centroids.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    data : matrix-like
        Data matrix of shape (num_inst, dim)
    centroids : matrix-like
        Centroids (centers) of shape (num_centroids, dim)

    Returns
    -------
    float
        Social Fairness Ratio

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import social_fairness_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> data = np.array(
    ...     [[-1, 1],
    ...      [1, 1],
    ...      [1, 1],
    ...      [0, -1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [1, 1],
    ...      [0, -1]])
    >>> centroids = np.array(
    ...     [[-2, 1],
    ...      [1, 2],
    ...      [0, -2]])
    >>> social_fairness_ratio(group_a, group_b, data, centroids)
    1.0
    """
    # check and coerce inputs
    group_a, group_b, _, _, data, centroids = _clustering_checks(
        group_a, group_b, y_pred=None, y_true=None, data=data, centroids=centroids
    )

    # Split by group
    data_a = data[group_a == 1]
    data_b = data[group_b == 1]

    # Reshape matrices for vectorization
    data_a = data_a.reshape(data_a.shape[0], 1, data_a.shape[1])
    data_b = data_b.reshape(data_b.shape[0], 1, data_b.shape[1])
    centroids = centroids.reshape(1, centroids.shape[0], centroids.shape[1])

    # Calculate distances
    dist_a = np.sqrt(((data_a - centroids) ** 2).sum(axis=-1))
    dist_b = np.sqrt(((data_b - centroids) ** 2).sum(axis=-1))

    # Take minimum over centroids and average over instances
    dist_a = dist_a.min(axis=1).mean(axis=0)
    dist_b = dist_b.min(axis=1).mean(axis=0)

    # return ratio of averages
    return dist_a / dist_b


def silhouette_diff(group_a, group_b, data, y_pred):
    r"""
    Silhouette Difference.

    Description
    -----------
    We compute the difference of the mean silhouette score for both
    groups.

    Interpretation
    --------------
    The silhouette difference ranges from -1 to 1, with lower values indicating bias
    towards group_a and larger values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)
    data : matrix-like
        Data matrix of shape (num_inst, dim)

    Returns
    -------
    float
        Silhouette difference : :math:`\texttt{mean_silhouette_a - mean_silhouette_b}`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import silhouette_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> data = np.array(
    ...     [[-1, 1],
    ...      [1, 1],
    ...      [1, 1],
    ...      [0, -1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [-1, 1],
    ...      [1, 1],
    ...      [0, -1]])
    >>> y_pred = np.array([0, 1, 1, 2, 0, 0, 0, 0, 1, 2])
    >>> silhouette_diff(group_a, group_b, data, y_pred)
    0.0
    """
    # check and coerce inputs
    group_a, group_b, y_pred, _, data, _ = _clustering_checks(
        group_a, group_b, y_pred, data=data
    )

    # Compute silhouette scores
    scores = silhouette_samples(data, y_pred, metric="euclidean")

    # Split min and maj
    scores_a = scores[group_a == 1]
    scores_b = scores[group_b == 1]

    return 0.5 * (np.mean(scores_a) - np.mean(scores_b))


def clustering_bias_metrics(
    group_a,
    group_b,
    y_pred,
    data=None,
    centroids=None,
    metric_type="equal_outcome",
):
    """
    Clustering bias metrics batch computation.

    Description
    -----------
    This function computes all the relevant clustering bias metrics,
    and displays them as a pandas dataframe.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Cluster predictions (categorical)
    data (optional) : matrix-like
        Data matrix of shape (num_inst, dim)
    centroids (optional) : matrix-like
        Centroids (centers)
    metric_type : 'both', 'equal_outcome' or 'equal_opportunity'
        Specifies which metrics we compute

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference
    """
    perform = {
        "Cluster Balance": cluster_balance,
        "Minimum Cluster Ratio": min_cluster_ratio,
        "Cluster Distribution Total Variation": cluster_dist_l1,
        "Cluster Distribution KL Div": cluster_dist_kl,
    }

    centroids_data_perform = {"Social Fairness Ratio": social_fairness_ratio}

    pred_data_perform = {"Silhouette Difference": silhouette_diff}

    ref_vals = {
        "Cluster Balance": 1,
        "Minimum Cluster Ratio": 1,
        "Cluster Distribution Total Variation": 0,
        "Cluster Distribution KL Div": 0,
        "Social Fairness Ratio": 1,
        "Silhouette Difference": 0,
        "Mutual Information Difference": 0,
    }

    out_metrics = []
    opp_metrics = []

    out_metrics += [
        [pf, fn(group_a, group_b, y_pred), ref_vals[pf]] for pf, fn in perform.items()
    ]
    if data is not None and centroids is not None:
        out_metrics += [
            [pf, fn(group_a, group_b, data, centroids), ref_vals[pf]]
            for pf, fn in centroids_data_perform.items()
        ]
    if data is not None and y_pred is not None:
        out_metrics += [
            [pf, fn(group_a, group_b, data, y_pred), ref_vals[pf]]
            for pf, fn in pred_data_perform.items()
        ]

    if metric_type == "both":
        metrics = out_metrics + opp_metrics
        return pd.DataFrame(
            metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_outcome":
        return pd.DataFrame(
            out_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_opportunity":
        return pd.DataFrame(
            opp_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    else:
        raise ValueError(
            "metric_type is not one of : both, equal_outcome, equal_opportunity"
        )



import numpy as np
import pandas as pd

# from holisticai.bias.mitigation.commons.disparate_impact_remover.algorithm import GroupRepairer
from holisticai.bias.mitigation.preprocessing.disparate_impact_remover import (
    DisparateImpactRemover,
)
from holisticai.utils.transformers.bias import BMPostprocessing as BMPos
from holisticai.utils.transformers.bias import SensitiveGroups


class DisparateImpactRemoverRS(BMPos):
    """
    Disparate impact remover edits feature values to increase group fairness
    while preserving rank-ordering within groups.

    References
    ----------
        [1] Feldman, Michael, et al. "Certifying and removing disparate impact."
        proceedings of the 21th ACM SIGKDD international conference on knowledge
        discovery and data mining. 2015.
    """

    def __init__(
        self,
        query_col="X",
        group_col="protected",
        score_col="score",
        repair_level=1.0,
        verbose=0,
    ):
        """
        Description
        -----------
        Init Disparate Exposure in Learning To Rank class.

        Parameters
        ----------
        group_col: str
            Name of the column in data that contains protected attribute.

        score_col:
            Name of the column in data that contains judgment values.

        repair_level : float
            Repair amount 0.0 (min) -> 1.0 (max)

        verbose : int
            If > 0, print progress.
        """
        # assign mandatory parameters
        self.score_col = score_col
        self.query_col = query_col
        self.group_col = group_col
        self.verbose = verbose
        self._assert_parameters(repair_level)
        self.repair_level = repair_level
        self.sensgroup = SensitiveGroups()
        self.dir = DisparateImpactRemover(repair_level=repair_level)

    def _assert_parameters(self, repair_level):
        if not 0.0 <= repair_level <= 1.0:
            raise ValueError("'repair_level' must be between 0.0 and 1.0.")

    def _filter_invalid_examples(self, rankings):
        new_rankings = []
        for q, ranking in rankings.groupby(self.query_col):
            if (ranking[self.group_col].sum() > 0).any():
                new_rankings.append(ranking)
        new_rankings = pd.concat(new_rankings, axis=0).reset_index(drop=True)
        return new_rankings

    def fit_transform(self, rankings: pd.DataFrame):
        return self.fit().transform(rankings)

    def fit(self):
        return self

    def transform(self, rankings: pd.DataFrame):
        """
        Description
        -----------
        Train a Disparate Exposure model to rank the prediction set.

        Parameters
        ----------
        rankings:    DataFrame

        Return
        ------
            DataFrame
        """
        rankings = self._filter_invalid_examples(rankings)
        new_rankings = [
            self.transform_features(ranking)
            for q, ranking in rankings.groupby(self.query_col)
        ]

        return pd.concat(new_rankings, axis=0).reset_index(drop=True)

    def transform_features(self, ranking: pd.DataFrame):
        """
        Transform data

        Description
        -----------
        Transform data to a fair representation

        Parameters
        ----------
        ranking : DataFrame
            Input data

        Return
        ------
            Self
        """
        data = np.c_[ranking[self.score_col].to_numpy()].tolist()
        group_a = ranking[self.group_col] == True
        group_b = ranking[self.group_col] == False

        new_data_matrix = self.dir.fit_transform(data, group_a, group_b)
        new_ranking = ranking.copy()
        new_ranking[self.score_col] = new_data_matrix
        new_ranking = new_ranking.reset_index(drop=True)
        new_ranking = (
            new_ranking.groupby(self.query_col)
            .apply(lambda df: df.sort_values(self.score_col, ascending=False))
            .reset_index(drop=True)
        )
        return new_ranking



from typing import Optional

import numpy as np
import pandas as pd

from .item_selection.selectors import ConventionalItemsSelection


class RecommenderSystemBase:
    """
    Recommender System Base Transformer
    """

    def predict(self, X: Optional[np.ndarray], top_n: Optional[int] = 10):

        self.item_selection = ConventionalItemsSelection(top_n=top_n)
        time_mask = X > 0
        selected_items = self.item_selection(time_mask, self.pred)[0]
        dfs = []
        for query_id, indexes in enumerate(selected_items):
            df = pd.DataFrame()
            df["score"] = np.array(self.pred[query_id, indexes])
            df["Y"] = np.array(indexes)
            df["X"] = query_id
            dfs.append(df)
        dfs = pd.concat(dfs, axis=0)
        return dfs[["X", "Y", "score"]]



from typing import Optional

import numpy as np
from sklearn.metrics._classification import _check_zero_division, _prf_divide

from holisticai.utils.transformers.bias import BMPostprocessing as BMPost


class CalibratedEqualizedOdds(BMPost):
    """
    Calibrated equalized odds postprocessing optimizes over calibrated classifier score outputs to find
    probabilities with which to change output labels with an equalized odds objective.
    References:
        Pleiss, Geoff, et al. "On fairness and calibration."
        Advances in neural information processing systems 30 (2017).
    """

    COST_CONSTRAINT = ["fnr", "fpr", "weighted"]

    def __init__(
        self,
        cost_constraint: Optional[str] = "fnr",
        alpha: Optional[float] = None,
    ):
        """
        Create a Calibrated Equalized Odds Post-processing instance.

        Parameters
        ----------
        cost_constraint : str
            Strategy used to evalute the cost function  The available contraints  are:
            [
                "fnr",
                "fpr",
                "weighted"
            ]
            false negative rate (fnr), false positive rate (fpr), and weighted

        alpha : float
            Used only with cost contraint  "weighted".
            Value between 0 and 1 used to combine fnr and fpr cost constraint.
        """
        self.cost_constraint = cost_constraint
        self.alpha = alpha

    def _build_cost(self, y_true, y_score, base_rate, sample_weight):
        if self.cost_constraint == "fpr":
            alpha = 0
        elif self.cost_constraint == "fnr":
            alpha = 1
        elif self.cost_constraint == "weighted":
            alpha = base_rate if self.alpha is None else self.alpha
        else:
            raise ValueError(f"unknown cost constraint: {self.cost_constraint}")

        gfpr = generalized_fpr(y_true, y_score, sample_weight=sample_weight)
        gfnr = generalized_fnr(y_true, y_score, sample_weight=sample_weight)

        return (1 - alpha) * gfpr + alpha * gfnr

    def _build_cost_variables(self, y_true, y_score, sample_weight):
        base_rate = y_true.mean()
        build_cost = lambda score: self._build_cost(
            y_true=y_true,
            y_score=score,
            base_rate=base_rate,
            sample_weight=sample_weight,
        )

        cost = build_cost(y_score)
        trivial_cost = build_cost(np.full_like(y_score, base_rate))
        return base_rate, cost, trivial_cost

    def _mitigate_bias_score(self, y_score, group, mix_rate, base_rate):
        indexes = np.random.random(sum(group)) <= mix_rate
        new_y_score = y_score[group == 1].copy()
        new_y_score[indexes] = base_rate
        return new_y_score

    def fit(
        self,
        y_true: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
    ):
        """
        Compute parameters for calibrated equalized odds.

        Description
        ----------
        Compute parameters for calibrated equalized odds algorithm.

        Parameters
        ----------
        y_true : array-like
            Target vector
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        sample_weight : array-like
            Sample weights. Used to weight generalized false positive rate (GFPR) and
            generalized false negative rate (GFNR). Sample weights could be used during training
            or computed by previous preprocessing strategy.
        Returns
        -------
        Self
        """
        params = self._load_data(
            y_true=y_true,
            y_proba=y_proba,
            group_a=group_a,
            group_b=group_b,
            sample_weight=sample_weight,
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_true = params["y_true"]
        y_score = params["y_score"]
        sample_weight = params["sample_weight"]

        self.base_rate_a, a_cost, a_trivial_cost = self._build_cost_variables(
            y_true[group_a], y_score[group_a], sample_weight=sample_weight[group_a]
        )

        self.base_rate_b, b_cost, b_trivial_cost = self._build_cost_variables(
            y_true[group_b], y_score[group_b], sample_weight=sample_weight[group_b]
        )

        b_costs_more = b_cost > a_cost
        self.a_mix_rate = (
            (b_cost - a_cost) / (a_trivial_cost - a_cost) if b_costs_more else 0
        )
        self.b_mix_rate = (
            0 if b_costs_more else (a_cost - b_cost) / (b_trivial_cost - b_cost)
        )

        return self

    def transform(
        self,
        y_pred: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        threshold: Optional[float] = 0.5,
    ):
        """
        Apply transform function to predictions and likelihoods

        Description
        ----------
        Use a fitted probability to change the output label and invert the likelihood

        Parameters
        ----------
        y_pred : array-like
            Predicted vector (nb_examlpes,)
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        threshold : float
            float value to discriminate between 0 and 1

        Returns
        -------
        dictionnary with new predictions
        """

        params = self._load_data(
            y_pred=y_pred, y_proba=y_proba, group_a=group_a, group_b=group_b
        )

        group_a = params["group_a"] == 1
        group_b = params["group_b"] == 1
        y_pred = params["y_pred"]
        y_score = params["y_score"]

        new_y_score_a = self._mitigate_bias_score(
            y_score, group_a, self.a_mix_rate, self.base_rate_a
        )
        new_y_score_b = self._mitigate_bias_score(
            y_score, group_b, self.b_mix_rate, self.base_rate_b
        )

        new_y_score = y_score.copy()
        new_y_score[group_a] = new_y_score_a
        new_y_score[group_b] = new_y_score_b

        new_y_pred = y_pred.copy()
        new_y_pred[group_a] = np.where(new_y_score[group_a] >= threshold, 1, 0)
        new_y_pred[group_b] = np.where(new_y_score[group_b] >= threshold, 1, 0)

        return {
            "y_pred": new_y_pred,
            "y_score": new_y_score,
        }

    def fit_transform(
        self,
        y_true: np.ndarray,
        y_proba: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
        sample_weight: Optional[np.ndarray] = None,
        threshold: Optional[float] = 0.5,
    ):
        """
        Fit and transform

        Description
        ----------
        Fit and transform

        Parameters
        ----------
        y_true : array-like
            Target vector
        y_proba : matrix-like
            Predicted probability matrix (num_examples, num_classes). The probability
            estimates must sum to 1 across the possible classes and each matrix value
            must be in the interval [0,1].
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)
        sample_weight : array-like
            Sample weights. Used to weight generalized false positive rate (GFPR) and
            generalized false negative rate (GFNR). Sample weights could be used during training
            or computed by previous preprocessing strategy.

        threshold : float
            float value to discriminate between 0 and 1

        Returns
        -------
        dictionnary with new predictions
        """
        return self.fit(
            y_true,
            y_proba,
            group_a,
            group_b,
            sample_weight,
        ).transform(y_true, y_proba, group_a, group_b, threshold)


def generalized_fpr(y_true, y_score, sample_weight):
    neg_idx = y_true != 1
    neg_weights = sample_weight[neg_idx]
    neg = neg_weights.sum()
    fpr = (y_score[neg_idx] * neg_weights).sum() / neg
    return fpr


def generalized_fnr(y_true, y_score, sample_weight):
    pos_idx = y_true == 1
    pos_weights = sample_weight[pos_idx]
    pos = pos_weights.sum()
    fnr = ((1 - y_score)[pos_idx] * pos_weights).sum() / pos
    return fnr



# Base Imports
import numpy as np
import pandas as pd

# Efficacy metrics
from sklearn.metrics import accuracy_score, confusion_matrix, roc_auc_score

# utils
from ...utils._validation import (
    _check_binary,
    _classification_checks,
    _regression_checks,
)


def _group_success_rate(g, y):
    """
    Group success rate.

    Description
    -----------
    This function computes the success rate for a given subgroup.

    Parameters
    ----------
    g : array-like
        subgroup vector (binary)
    y : array-like
        predictions vector (binary)

    Returns
    -------
    float
        group success rate

    """
    success_rate = y[g == 1].sum() / g.sum()  # success rate group_a
    return success_rate


def statistical_parity(group_a, group_b, y_pred):
    """
    Statistical parity.

    Description
    -----------
    This function computes the statistical parity (difference of success rates)
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. Negative values are unfair towards group_a.
    Positive values are unfair towards group_b. The range (-0.1,0.1)
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Statistical Parity : :math:`sr_a - sr_b`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import statistical_parity
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> statistical_parity(group_a, group_b, y_pred)
    0.4166666666666667
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return sr_a - sr_b


def success_rate(group_a, group_b, y_pred):
    """
    Calculates the raw success rates for each group.
    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    Returns
    -------
    dict
        Dictionary with two keys, sr_a and sr_b (success rate for group a and b)
    """
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    return {"sr_a": sr_a, "sr_b": sr_b}


def disparate_impact(group_a, group_b, y_pred):
    """
    Disparate Impact.

    Description
    -----------
    This function computes the disparate impact (ratio of success rates)
    between group_a and group_b class.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 are unfair towards group_a.
    Values above 1 are unfair towards group_b. The range (0.8,1.2)
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Disparate Impact : :math:`sr_a/sr_b`

    References
    ----------
    .. [1] `M B Zafar, I Valera, M G Rodriguez, K P. Gummadi (2017).
            Fairness Constraints: Mechanisms for Fair Classification, MPI-SWS
            <https://arxiv.org/pdf/1507.05259.pdf>`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import disparate_impact
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> disparate_impact(group_a, group_b, y_pred)
    2.25
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return sr_a / sr_b


def four_fifths(group_a, group_b, y_pred):
    """
    Four Fifths.

    Description
    -----------
    This function computes the four fifths rule (ratio of success rates)
    between group_a and group_b. We return the minimum of the ratio
    taken both ways.

    Interpretation
    --------------
    A value of 1 is desired. Values below 1 are unfair. The range (0.8,1)
    is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Four Fifths

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import four_fifths
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> four_fifths(group_a, group_b, y_pred)
    0.4444444444444444
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    return min(sr_a / sr_b, sr_b / sr_a)


def cohen_d(group_a, group_b, y_pred):
    r"""
    Cohen D.

    Description
    -----------
    This function computes the Cohen D statistic (normalised statistical parity)
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. Negative values are unfair towards group_a.
    Positive values are unfair towards group_b. Reference values:
    0.2 is considered a small effect size, 0.5 is considered medium,
    0.8 is considered large.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Cohen D : :math:`\frac{sr_a-sr_b}{\texttt{std_pool}}`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import cohen_d
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> cohen_d(group_a, group_b, y_pred)
    -0.7844645405527363
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b

    # calculate STD_a and STD_b
    std_b = np.sqrt(sr_b * (1 - sr_b))
    std_a = np.sqrt(sr_a * (1 - sr_a))

    # calculate n_a and n_b
    n_a = group_a.sum()
    n_b = group_b.sum()

    # calculate poolSTD
    std_pool = np.sqrt(
        ((n_b - 1) * std_b**2 + (n_a - 1) * std_a**2) / (n_a + n_b - 2)
    )

    return (sr_a - sr_b) / std_pool


def z_test_diff(group_a, group_b, y_pred):
    r"""
    Z Test (Difference).

    Description
    -----------
    This function computes the Z-test statistic for the difference
    in success rates. Also known as 2-SD Statistic.

    Interpretation
    --------------
    A value of 0 is desired. This test considers the data unfair
    if the computed value is greater than 2 or smaller than -2,
    indicating a statistically significant difference in success rates.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Z test (difference version)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import z_test_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> z_test_diff(group_a, group_b, y_pred)
    1.290994449

    References
    ----------
    .. [1] `Morris (2001).
           Sample size requirements for adverse impact analysis
            <https://www.semanticscholar.org/paper/Sample-Size-Required-for-Adverse-Impact-Analysis-Morris/877f7acd7c646a21f4947166a07f41664dcabe95>`

    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    n_a = group_a.sum()
    n_b = group_b.sum()
    sr_tot = (sr_a * n_a + sr_b * n_b) / (n_a + n_b)
    n_tot = n_a + n_b

    # calculate p_a
    p_a = n_a / n_tot

    return (sr_a - sr_b) / np.sqrt((sr_tot * (1 - sr_tot)) / (n_tot * p_a * (1 - p_a)))


def z_test_ratio(group_a, group_b, y_pred):
    r"""
    Z Test (Ratio).

    Description
    -----------
    This function computes the Z-test statistic for the ratio
    in success rates. Also known as 2-SD Statistic.

    Interpretation
    --------------
    A value of 0 is desired. This test considers the data unfair
    if the computed value is greater than 2 or smaller than -2,
    indicating a statistically significant ratio in success rates.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)

    Returns
    -------
    float
        Z-test (ratio version)

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import z_test_ratio
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 1, 0, 1, 1, 0, 0, 0, 0])
    >>> z_test_ratio(group_a, group_b, y_pred)
    1.256287689

    References
    ----------
    .. [1] `Morris (2001).
           Sample size requirements for adverse impact analysis
            <https://www.semanticscholar.org/paper/Sample-Size-Required-for-Adverse-Impact-Analysis-Morris/877f7acd7c646a21f4947166a07f41664dcabe95>`
    """
    # check and coerce
    group_a, group_b, y_pred, _ = _classification_checks(
        group_a, group_b, y_pred, y_true=None
    )

    # calculate sr_a and sr_b
    sr_a = _group_success_rate(group_a, y_pred)  # success rate group_a
    sr_b = _group_success_rate(group_b, y_pred)  # success rate group_b
    n_a = group_a.sum()
    n_b = group_b.sum()
    sr_tot = (sr_a * n_a + sr_b * n_b) / (n_a + n_b)
    n_tot = n_a + n_b

    # calculate p_a
    p_a = n_a / n_tot

    return (np.log(sr_a / sr_b)) / np.sqrt(
        (1 - sr_tot) / (sr_tot * n_tot * p_a * (1 - p_a))
    )


def _correlation_diff(group_a, group_b, y_pred, y_true):
    """
    Correlation difference.

    Description
    -----------
    This function computes the difference in correlation between predicted
    and true labels for group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -2 and 2,
    with negative values indicating bias against group_a, and
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        Correlation Difference : :math:`CV_a - CV_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import correlation_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> correlation_diff(group_a, group_b, y_pred, y_true)
    1.4472135954999579
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Calculate Pearson correlations
    cv_a = np.corrcoef(y_pred[group_a == 1], y_true[group_a == 1])[1, 0]
    cv_b = np.corrcoef(y_pred[group_b == 1], y_true[group_b == 1])[1, 0]

    return cv_a - cv_b


def equal_opportunity_diff(group_a, group_b, y_pred, y_true):
    """
    Equality of opportunity difference.

    Description
    -----------
    This function computes the difference in true positive
    rates for group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a, and
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        Equal opportunity difference : :math:`tpr_a - tpr_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import equal_opportunity_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> equal_opportunity_diff(group_a, group_b, y_pred, y_true)
    0.33333333333333337
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Calculate true positive rates
    tpr_a = confusion_matrix(
        y_true[group_a == 1], y_pred[group_a == 1], normalize="true"
    )[1, 1]
    tpr_b = confusion_matrix(
        y_true[group_b == 1], y_pred[group_b == 1], normalize="true"
    )[1, 1]

    return tpr_a - tpr_b


def false_positive_rate_diff(group_a, group_b, y_pred, y_true):
    """
    False positive rate difference.

    Description
    ----------
    This function computes the difference in false positive
    rates between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a, and
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        FPR_diff : :math:`fpr_a - fpr_b`

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import false_positive_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> false_positive_diff(group_a, group_b, y_pred, y_true)
    -1.0
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Calculate false positive rates
    fpr_a = confusion_matrix(
        y_true[group_a == 1], y_pred[group_a == 1], normalize="true"
    )[0, 1]
    fpr_b = confusion_matrix(
        y_true[group_b == 1], y_pred[group_b == 1], normalize="true"
    )[0, 1]

    return fpr_a - fpr_b


def false_negative_rate_diff(group_a, group_b, y_pred, y_true):
    """
    False negative Rate difference.

    Description
    ----------
    This function computes the difference in false negative
    rates for group_a and group_b.

    Interpretation
    ----------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_b, and
    positive values indicating bias against group_a.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        False Negative Rate difference : :math:`fnr_a - fnr_b`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import fnr_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> fnr_diff(group_a, group_b, y_pred, y_true)
    -0.3333333333333333
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Calculate false negative rates
    fnr_a = confusion_matrix(
        y_true[group_a == 1], y_pred[group_a == 1], normalize="true"
    )[1, 0]
    fnr_b = confusion_matrix(
        y_true[group_b == 1], y_pred[group_b == 1], normalize="true"
    )[1, 0]

    return fnr_a - fnr_b


def true_negative_rate_diff(group_a, group_b, y_pred, y_true):
    """
    True negative Rate difference.

    Description
    ----------
    This function computes the difference in true negative
    rates for group_a and group_b.

    Interpretation
    ----------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a, and
    positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        True Negative Rate difference : :math:`tnr_a - tnr_b`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import tnr_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> tnr_diff(group_a, group_b, y_pred, y_true)
    1
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Calculate false negative rates
    tnr_a = confusion_matrix(
        y_true[group_a == 1], y_pred[group_a == 1], normalize="true"
    )[0, 0]
    tnr_b = confusion_matrix(
        y_true[group_b == 1], y_pred[group_b == 1], normalize="true"
    )[0, 0]

    return tnr_a - tnr_b


def average_odds_diff(group_a, group_b, y_pred, y_true):
    """
    Average Odds Difference.

    Description
    -----------
    This function computes the difference in average odds
    between group_a and group_b.

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a,
    and positive values indicating bias against group_b.
    The range (-0.1,0.1) is considered acceptable.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        AOD : :math:`0.5 * (fpr_a-fpr_b + tpr_a-tpr_b)`

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import average_odds_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> average_odds_diff(group_a, group_b, y_pred, y_true)
    -0.3333333333333333
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(
        group_a, group_b, y_pred, y_true
    )

    # Compute AOD
    aod = 0.5 * (
        equal_opportunity_diff(group_a, group_b, y_pred, y_true)
        + false_positive_rate_diff(group_a, group_b, y_pred, y_true)
    )

    return aod


def accuracy_diff(group_a, group_b, y_score, y_true):
    """
    Accuracy Difference.

    Description
    ----------
    This function computes the difference in accuracy
    of predictions for group_a and group_b

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a,
    and positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_score : numpy array
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        acc_diff : acc_a - acc_b

    Examples
    -------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import accuracy_diff
    >>> group_a = np.array([1, 1, 1, 1, 0, 0, 0, 0, 0, 0])
    >>> group_b = np.array([0, 0, 0, 0, 1, 1, 1, 1, 1, 1])
    >>> y_pred = np.array([1, 1, 0, 0, 1, 1, 0, 1, 1, 1])
    >>> y_true = np.array([1, 1, 0, 0, 1, 0, 1, 0, 0, 1])
    >>> accuracy_diff(group_a, group_b, y_pred, y_true)
    0.6666666666666667
    """
    # check and coerce
    group_a, group_b, y_score, y_true, _ = _regression_checks(
        group_a, group_b, y_score, y_true, None
    )
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_score_a = y_score[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_score_b = y_score[group_b == 1]

    # compute abroca
    val = accuracy_score(y_true_a, y_score_a) - accuracy_score(y_true_b, y_score_b)
    return val


def abroca(group_a, group_b, y_score, y_true):
    """
    ABROCA (area between roc curves).

    Description
    -----------
    This function computes the area between the roc curve
    of group_a and the roc curve of group_b

    Interpretation
    --------------
    A value of 0 is desired. This metric ranges between -1 and 1,
    with negative values indicating bias against group_a,
    and positive values indicating bias against group_b.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_score : array-like
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)

    Returns
    -------
    float
        ABROCA : roc_auc_a - roc_auc_b

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.bias.metrics import abroca
    >>> group_a = np.array([1]*50 + [0]*50)
    >>> group_b = np.array([0]*50 + [1]*50)
    >>> y_score = np.concatenate((np.linspace(0,1,50),np.linspace(0,1,50)**2))
    >>> y_true = y_score + np.random.random(y_score.shape)>0.5
    >>> abroca(group_a, group_b, y_score, y_true)
    0.11806478405315601
    """
    # check and coerce
    group_a, group_b, y_score, y_true, _ = _regression_checks(
        group_a, group_b, y_score, y_true, None
    )
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_score_a = y_score[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_score_b = y_score[group_b == 1]

    # compute abroca
    val = roc_auc_score(y_true_a, y_score_a) - roc_auc_score(y_true_b, y_score_b)
    return val


def classification_bias_metrics(
    group_a, group_b, y_pred, y_true=None, y_score=None, metric_type="equal_outcome"
):
    """
    Classification bias metrics batch computation.

    Description
    -----------
    This function computes all the relevant classification bias metrics,
    and displays them as a pandas dataframe. It also includes a fair reference
    value for comparison.

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Predictions vector (binary)
    y_true (optional) : numpy array
        Target vector (binary)
    y_score (optional) : array-like
        Probability estimates (regression)
    metric_type (optional) : 'both', 'equal_outcome' or 'equal_opportunity'
        Specifies which metrics we compute

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference

    Examples
    --------
    """
    equal_outcome_metrics = {
        "Statistical Parity": statistical_parity,
        "Disparate Impact": disparate_impact,
        "Four Fifths Rule": four_fifths,
        "Cohen D": cohen_d,
        "2SD Rule": z_test_diff,
    }

    equal_opportunity_metrics = {
        "Equality of Opportunity Difference": equal_opportunity_diff,
        "False Positive Rate Difference": false_positive_rate_diff,
        "Average Odds Difference": average_odds_diff,
        "Accuracy Difference": accuracy_diff,
    }

    soft_metrics = {
        "ABROCA": abroca,
    }

    ref_vals = {
        "Statistical Parity": 0,
        "Disparate Impact": 1,
        "Four Fifths Rule": 1,
        "Cohen D": 0,
        "Equality of Opportunity Difference": 0,
        "False Positive Rate Difference": 0,
        "Average Odds Difference": 0,
        "Accuracy Difference": 0,
        "ABROCA": 0,
        "2SD Rule": 0,
    }

    out_metrics = [
        [pf, fn(group_a, group_b, y_pred), ref_vals[pf]]
        for pf, fn in equal_outcome_metrics.items()
    ]
    if y_true is not None:
        opp_metrics = [
            [pf, fn(group_a, group_b, y_pred, y_true), ref_vals[pf]]
            for pf, fn in equal_opportunity_metrics.items()
        ]
    if y_score is not None:
        opp_metrics += [
            [pf, fn(group_a, group_b, y_score, y_true), ref_vals[pf]]
            for pf, fn in soft_metrics.items()
        ]

    if metric_type == "both":
        metrics = out_metrics + opp_metrics
        return pd.DataFrame(
            metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_outcome":
        return pd.DataFrame(
            out_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    elif metric_type == "equal_opportunity":
        return pd.DataFrame(
            opp_metrics, columns=["Metric", "Value", "Reference"]
        ).set_index("Metric")

    else:
        raise ValueError(
            "metric_type is not one of : both, equal_outcome, equal_opportunity"
        )



# Base Imports
import numpy as np


def get_colors(n_lines, extended_colors=False, reverse=False):
    """
    Get colors for Holistic AI pallette

    Description
    ----------
    This function returns an array of colors
    for plotting using the Holistic AI pallette.

    Parameters
    ----------
    n_lines : scalar
        Number of lines

    Returns
    -------
    List of colors in RGB format
    """

    purple = np.array([161 / 255, 103 / 255, 211 / 255])
    blue = np.array([68 / 255, 168 / 255, 249 / 255])

    colors = [purple + t * (blue - purple) for t in np.linspace(0, 1, n_lines)]

    if extended_colors and n_lines > 3:
        n_lines_extremes = round(n_lines / 3) + 1
        n_lines_middle = n_lines - 2 * n_lines_extremes + 2

        purple_to_blue = [
            purple + t * (blue - purple) for t in np.linspace(0, 1, n_lines_middle)
        ]

        white = np.ones(blue.shape) * 0.8
        black = np.zeros(blue.shape) + 0.3

        black_to_purple = [
            black + t * (purple - black) for t in np.linspace(0, 1, n_lines_extremes)
        ]
        blue_to_white = [
            blue + t * (white - blue) for t in np.linspace(0, 1, n_lines_extremes)
        ]

        colors = black_to_purple[:-1] + purple_to_blue + blue_to_white[1:]

    if reverse:
        colors = colors[::-1]

    return colors



import numpy as np
import pandas as pd

from .._conventions import _ALL, _EVENT, _GROUP_ID, _LABEL, _LOSS, _PREDICTION, _SIGNED
from .._moments_utils import BaseMoment, format_data


class RegressionConstraint(BaseMoment):
    """Constrain the mean loss or the worst-case loss by a group"""

    PROBLEM_TYPE = "regression"

    def __init__(self, loss, upper_bound=None, no_groups=False):
        """
        Parameters
        ----------
        loss : {SquareLoss, AbsoluteLoss}
            A loss object with an `eval` method, e.g. `SquareLoss` or
            `AbsoluteLoss`.
        upper_bound : float
            An upper bound on the loss, also referred to as :math:`\\zeta`;
            `upper_bound` is an optional argument that is not always
            required; default None
        no_groups : bool
            indicates whether to calculate the mean loss or group-level losses,
            default False, i.e., group-level losses are the default behavior
        """
        super().__init__()
        self.reduction_loss = loss
        self.upper_bound = upper_bound
        self.no_groups = no_groups

    def default_objective(self):
        """Return a default objective."""
        return MeanLoss(self.reduction_loss)

    def load_data(self, X, y, sensitive_features):
        params = format_data(y=y)
        y = params["y"]

        if self.no_groups:
            sensitive_features = np.zeros_like(y)

        # The following uses X and not X_train so that the estimators get X untouched
        super().load_data(X, y, sensitive_features=sensitive_features)
        event = pd.Series(data=_ALL, index=self.tags[_LABEL].index)
        self.tags[_EVENT] = event
        self.event_ids = np.sort(self.tags[_EVENT].dropna().unique())
        self.event_prob = self.tags[_EVENT].dropna().value_counts() / len(self.tags)

        self.group_prob = (
            self.tags.groupby(_GROUP_ID).size() / self.total_samples
        )  # self.tags[_GROUP_ID].dropna().value_counts() / len(self.tags)
        self.group_values = np.sort(self.tags[_GROUP_ID].unique())

        self.index = self.group_prob.index
        self.default_objective_lambda_vec = self.group_prob

        self._get_basis()

    def _get_basis(self):
        pos_basis = pd.DataFrame()
        neg_basis = pd.DataFrame()
        neg_basis_present = pd.Series(dtype="float64")
        zero_vec = pd.Series(0.0, self.index)
        i = 0
        for group in self.group_values:
            pos_basis[i] = zero_vec
            neg_basis[i] = zero_vec
            pos_basis[i][group] = 1
            neg_basis_present.at[i] = False
            i += 1

        self.neg_basis_present = neg_basis_present
        self.basis = {"+": pos_basis, "-": neg_basis}
        self.default_objective_lambda_vec = self.group_prob

    def gamma(self, predictor):
        """Calculate the degree to which constraints are currently violated by the predictor."""
        self.tags[_PREDICTION] = predictor(self.X)
        self.tags[_LOSS] = self.reduction_loss.eval(
            self.tags[_LABEL], self.tags[_PREDICTION]
        )
        expect_attr = self.tags.groupby(_GROUP_ID).mean()
        return expect_attr[_LOSS]

    def bound(self):
        """Return the vector of bounds.
        Returns
        -------
        pandas.Series
            A vector of bounds on group-level losses
        """
        if self.upper_bound is None:
            raise ValueError("No Upper Bound")
        return pd.Series(self.upper_bound, index=self.index)

    def project_lambda(self, lambda_vec):
        """Return the lambda values."""
        return lambda_vec

    def signed_weights(self, lambda_vec=None):
        """Return the signed weights."""
        if lambda_vec is None:
            adjust = pd.Series(1.0, index=self.index)
        else:
            adjust = lambda_vec / self.group_prob
        return self.tags.apply(lambda row: adjust[row[_GROUP_ID]], axis=1)


class MeanLoss(RegressionConstraint):
    """
    Moment for evaluating the mean loss.
    """

    def __init__(self, loss):
        super().__init__(loss, upper_bound=None, no_groups=True)


class BoundedGroupLoss(RegressionConstraint):
    """
    Moment for constraining the worst-case loss by a group.
    """

    def __init__(self, loss, *, upper_bound=None):
        super().__init__(loss, upper_bound=upper_bound, no_groups=False)



from functools import partial

import numpy as np
from scipy.stats import multivariate_normal
from sklearn.metrics import accuracy_score

from holisticai.utils.transformers.bias import SensitiveGroups


class StatisticalRate:
    def __init__(self, mu):
        self.mu = mu
        self.num_params = 2

    def gamma(self, y_true, y_pred, groups_num):
        pos_0 = np.mean(y_pred[groups_num == 0] == 1)
        pos_1 = np.mean(y_pred[groups_num == 1] == 1)
        if pos_0 == 0 or pos_1 == 0:
            return 0
        return min(pos_0 / pos_1, pos_1 / pos_0)

    def forward(self, P, params, a=None, b=None, return_cs=False):

        l_1, l_2 = params
        z_0, z_1 = 1 - self.z_prior, self.z_prior

        c_0 = P["Y=1"] - 0.5
        c_1 = P["Z=0"] / z_0
        c_2 = P["Z=1"] / z_1

        t = c_0 + c_1 * l_1 + c_2 * l_2

        if return_cs:
            return t, c_1, c_2
        else:
            return t

    def _gradient(self, a, b, t, c, l):
        _t = t * c / np.sqrt(t**2 + self.mu**2)
        exp = np.mean(_t)
        dl = exp - b + (b - a) / 2 + (b - a) * l / (2 * np.sqrt(l**2 + self.mu**2))
        return dl

    def expected_gradient(self, P, params, a, b):
        l_1, l_2 = params
        t, c_1, c_2 = self.forward(P, params, return_cs=True)
        dl1 = self._gradient(a, b, t, c_1, l_1)
        dl2 = self._gradient(a, b, t, c_2, l_2)
        return dl1, dl2

    def cost_function(self, P, params, a, b):
        l_1, l_2 = params

        exp = np.mean(np.abs(self.forward(P, params)))
        result = exp - b * l_1 + -b * l_2
        if l_1 > 0:
            result += (b - a) * l_1
        if l_2 > 0:
            result += (b - a) * l_2

        return result


class FalseDiscovery:
    def __init__(self, mu):
        self.mu = mu
        self.num_params = 4

    def gamma(self, y_true, y_pred, groups_num):
        pos_0 = y_pred[groups_num == 0] == 1
        pos_1 = y_pred[groups_num == 1] == 1
        if np.sum(pos_0) == 0 or np.sum(pos_1) == 0:
            return 0
        fdr_0 = np.sum(pos_0 & (y_true[groups_num == 0] == -1)) / np.sum(pos_0)
        fdr_1 = np.sum(pos_1 & (y_true[groups_num == 1] == -1)) / np.sum(pos_1)
        if fdr_0 == 0 or fdr_1 == 0:
            return 0
        return min(fdr_0 / fdr_1, fdr_1 / fdr_0)

    def forward(self, P, params, a, b, return_probs=False):

        P["Y=-1|Z=0"] = P["Y=-1,Z=0"] / P["total"]
        P["Y=-1|Z=1"] = P["Y=-1,Z=1"] / P["total"]

        u_1, u_2, l_1, l_2 = params
        c_0 = P["Y=1"] - 0.5
        c_1 = u_1 * (P["Y=-1|Z=0"] - a * P["Z=0"]) + u_2 * (
            P["Y=-1|Z=1"] - a * P["Z=1"]
        )
        c_2 = l_1 * (-P["Y=-1|Z=0"] + b * P["Z=0"]) + l_2 * (
            -P["Y=-1|Z=1"] + b * P["Z=1"]
        )

        t = c_0 + c_1 * l_1 + c_2 * l_2

        if return_probs:
            return t, P
        else:
            return t

    def expected_gradient(self, P, params, a, b):
        t, P = self.forward(P, params, a, b, return_probs=True)
        res = np.vstack(
            [
                P["Y=-1|Z=0"] - a * P["Z=0"],
                P["Y=-1|Z=1"] - a * P["Z=1"],
                -P["Y=-1|Z=0"] + b * P["Z=0"],
                -P["Y=-1|Z=1"] + b * P["Z=1"],
            ]
        )
        res *= t / np.sqrt(t**2 + self.mu**2)
        return np.mean(res, axis=1)

    def cost_function(self, P, params, a, b):
        return np.mean(np.abs(self.forward(P, params, a, b)))



class ParametersHandler:
    def __init__(self, param_names=None, step_name=None):
        self.param_names = param_names
        self.step_name = step_name
        self.clean_parameters()

    def clean_parameters(self):
        self.dict_params = {}

    def set_shared_parameters(self, dict_params):
        self.dict_params = dict_params

    def __contains__(self, param_name):
        return param_name in self.dict_params

    def __getitem__(self, param_name):
        return self.dict_params[param_name]

    def __setitem__(self, param_name, param_value):
        self.dict_params[param_name] = param_value

    def feed(self, params, return_dropped=False):
        dropped_params = {}
        self.clean_parameters()
        if self.step_name:
            for name, value in params.items():
                if name.startswith(self.step_name):
                    param_name = name.split("__", 1)[1]
                    self[param_name] = value
                else:
                    dropped_params[name] = value

        elif self.param_names:

            for name, value in params.items():
                if name in self.param_names:
                    param_name = name.split("__", 1)[1]
                    self[param_name] = value
                else:
                    dropped_params[name] = value

        if return_dropped:
            return dropped_params


class PipelineParametersHandler:
    def __init__(self):
        self.bias_mitigator = ParametersHandler(
            param_names=["bm__group_a", "bm__group_b"]
        )

    def create_estimator_parameters(self, estimator_name, estimator):
        self.estimator = ParametersHandler(step_name=estimator_name)
        self.estimator_model = estimator

    def get_estimator_paramters(self):
        return self.estimator.dict_params



import numpy as np

# scikit
from sklearn.metrics import f1_score, precision_score, recall_score

# Formatting
from . import mat_to_binary


def avg_precision(mat_pred, mat_true, top=None, thresh=0.5):
    """
    Average Precision (recommender)

    Description
    ----------
    We average out the precision of recommender predictions over
    all users

    Parameters
    ----------
    mat_pred : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    mat_true : numpy ndarray
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each item.
    top : int
        If not None, the top k scores that are shown to each user
    thresh : float
        Float in (0,1) range indicating threshold at which
        a given item is shown to user

    Returns
    -------
    float
        Average precision
    """
    # Make matrices binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)
    binary_mat_true = mat_to_binary(mat_true, top=top, thresh=thresh)

    num_users, _ = mat_pred.shape
    vals = np.zeros((num_users,))

    # Sum accuracy over users
    for u in range(num_users):
        pred = binary_mat_pred[u]
        true = binary_mat_true[u]
        vals[u] = precision_score(true, pred)

    # Average
    return np.mean(vals)


def avg_recall(mat_pred, mat_true, top=None, thresh=0.5):
    """
    Average Recall (recommender)

    Description
    ----------
    We average out the recall of recommender predictions over
    all users

    Parameters
    ----------
    mat_pred : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    mat_true : numpy ndarray
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each item.
    top : int
        If not None, the top k scores that are shown to each user
    thresh : float
        Float in (0,1) range indicating threshold at which
        a given item is shown to user

    Returns
    -------
    float
        Average recall
    """
    # Make matrices binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)
    binary_mat_true = mat_to_binary(mat_true, top=top, thresh=thresh)

    num_users, _ = mat_pred.shape
    vals = np.zeros((num_users,))

    # Sum accuracy over users
    for u in range(num_users):
        pred = binary_mat_pred[u]
        true = binary_mat_true[u]
        vals[u] = recall_score(true, pred)

    # Average
    return np.mean(vals)


def avg_f1(mat_pred, mat_true, top=None, thresh=0.5):
    """
    Average f1 (recommender)

    Description
    ----------
    We average out the f1 of recommender predictions over
    all users

    Parameters
    ----------
    mat_pred : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    mat_true : numpy ndarray
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each item.
    top : int
        If not None, the top k scores that are shown to each user
    thresh : float
        Float in (0,1) range indicating threshold at which
        a given item is shown to user

    Returns
    -------
    float
        Average f1
    """
    # Make matrices binary
    binary_mat_pred = mat_to_binary(mat_pred, top=top, thresh=thresh)
    binary_mat_true = mat_to_binary(mat_true, top=top, thresh=thresh)

    num_users, _ = mat_pred.shape
    vals = np.zeros((num_users,))

    # Sum accuracy over users
    for u in range(num_users):
        pred = binary_mat_pred[u]
        true = binary_mat_true[u]
        vals[u] = f1_score(true, pred)

    # Average
    return np.mean(vals)


def recommender_rmse(mat_pred, mat_true):
    """
    Recommender RMSE

    Description
    ----------
    We compute the rmse of recommender predictions over
    all non null scores.

    Parameters
    ----------
    mat_pred : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    mat_true : numpy ndarray
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each item.

    Returns
    -------
    float
        Recommender RMSE
    """
    se_diff = (mat_pred - mat_true) ** 2
    rmse_val = np.sqrt(np.nanmean(se_diff))

    return rmse_val


def recommender_mae(mat_pred, mat_true):
    """
    Recommender MAE

    Description
    ----------
    We compute the mae of recommender predictions over
    all non null scores.

    Parameters
    ----------
    mat_pred : numpy ndarray
        Matrix with shape (num_users, num_items). A recommender
        score (binary or soft pred) for each item.
    mat_true : numpy ndarray
        Matrix with shape (num_users, num_items). A target score
        (binary or soft pred) for each item.

    Returns
    -------
    float
        Recommender MAE
    """
    abs_diff = np.abs(mat_pred - mat_true)
    mae_val = np.nanmean(abs_diff)

    return mae_val


def entropy(p, q=None):
    """
    Entropy

    Description
    ----------
    We compute the entropy of a probabvility or relative
    entropy of two probabilities

    Parameters
    ----------
    p : array
        probability vector
    q : array
        probability vector

    Returns
    -------
    float
        Entropy
    """
    p = np.array(p)
    p = p / np.sum(p)

    if q is None:
        return -np.sum(np.where(p != 0, p * np.log(p), 0))

    else:
        q = np.array(q)
        q = q / np.sum(q)
        return np.sum(np.where(p != 0, p * np.log(p / q), 0))
